/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.bus;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.channel.TCPIPProtocol;
import org.objectweb.dream.protocol.messagePassing.MessagePassingOutgoingPush;
import org.objectweb.dream.protocol.messagePassing.MessagePassingProtocol;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Local access manager specialized for lower TCP/IP protocol.
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class TCPIPLocalAccessManager 
    implements
      LocalAccessManager
{

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  @Requires(name=MessagePassingProtocol.LOWER_PROTOCOL_ITF_NAME)
  protected MessagePassingProtocol lowerLevelProtocolItf;

  // --------------------------------------------------------------------------
  // Services interfaces
  // --------------------------------------------------------------------------

  /**
   * Component reference
   */
  @Service
  Component                  weaveableC;

  /**
   * Logger of the component
   */
  @DreamMonolog()
  protected Logger           logger;
  
  // ---------------------------------------------------------------------------
  // Implementation of the LocalAccessManager interface
  // ---------------------------------------------------------------------------

  /**
   * @see LocalAccessManager#findLocalAddress(ExportIdentifier[])
   */
  public int findLocalAddress(ExportIdentifier[] exportIds)
      throws BindException
  {
    int address = -1;
    String hostName;
    try
    {
      hostName = InetAddress.getLocalHost().getHostName();
    }
    catch (UnknownHostException e)
    {
      throw new BindException("Can't find local address : " + e.getMessage());
    }
    for (int i = 0; i < exportIds.length; i++)
    {
      ExportIdentifier identifier = exportIds[i];
      if (identifier instanceof IPExportIdentifier)
      {
        IPExportIdentifier tcpIdentifier = (IPExportIdentifier) identifier;
        if (tcpIdentifier.getHostName().equals(hostName))
        {
          if (address != -1)
          {
            throw new BindException(
                "Ambiguous export identifier, can't determine local address");
          }
          address = i;
        }
      }
    }
    if (address == -1)
    {
      throw new BindException(
          "Can't find export identifier with the local address.");
    }
    return address;
  }

  /**
   * @see LocalAccessManager#exportBus(ExportIdentifier, IncomingPush)
   */
  public MessagePassingOutgoingPush exportBus(ExportIdentifier localId,
      IncomingPush incomingPush) throws ExportException
  {
    if (!(localId instanceof IPExportIdentifier))
    {
      throw new ExportException("invalid local export identifier.");
    }
    logger.log(BasicLevel.DEBUG, "Export local access to the bus");
    IPExportIdentifier ipIdentifier = (IPExportIdentifier) localId;
    Map<String, Object> hints = new HashMap<String, Object>();
    hints.put(TCPIPProtocol.PORT, Integer.valueOf(ipIdentifier.getPort()));
    hints.put(TCPIPProtocol.PORT_RANGE, Integer.valueOf(0));
    return lowerLevelProtocolItf.export(incomingPush, hints);
  }
}
