/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): Pierre GARCIA 
 */

package org.objectweb.dream.protocol.channel;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLServerSocketFactory;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManagerFactory;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.manager.TaskManager;
import org.objectweb.dream.control.activity.manager.ThreadPoolManager;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolOverflowException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamLifeCycle;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.dreamannotation.util.DreamLifeCycleType;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.message.codec.CodecInputOutput;
import org.objectweb.dream.message.codec.MessageCodec;
import org.objectweb.dream.message.codec.SocketCodecInputOutput;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExceptionChunk;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.dream.util.Error;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Implementation of the TCP/IP protocol.
 */
@DreamComponent(controllerDesc = "activeDreamUnstoppablePrimitive")
@Provides(interfaces = { @Interface(name = SSLProtocol.ITF_NAME, signature = SSLProtocol.class) })
public class SSLProtocolImpl implements SSLProtocol, ConnectionManager {
    /**
     * Variable for debug
     */
    private static final boolean DEFENSIVE_CHECK = true;

    /**
     * Name associated to the "bind-reply" chunk
     */
    protected static final String BIND_REPLY_CHUNK_NAME = "tcpip-bind-reply";

    // ------------------------------------------------------------------------
    // --
    // Services interfaces
    // ------------------------------------------------------------------------
    // --

    /**
     * Component reference
     */
    @Service
    Component weaveableC;

    /**
     * Logger of the component
     */
    @DreamMonolog()
    protected Logger logger;

    /**
     * Chunk factory used to create exception chunk. In the implementation of
     * the export/bind pattern, this chunk is used to notify that an exception
     * occurred while binding on the server side;
     */
    protected ChunkFactoryReference<ExceptionChunk> exceptionChunkFactory;

    /**
     * Set which contains all sessions
     */
    protected Set<Session> sessions = new HashSet<Session>();

    /**
     * Synchronized Queue which contains available sessions, i.e. sessions that
     * can be used by the reader task to receive messages.
     */
    protected BlockingQueue<Session> availableReadingSessions = new LinkedBlockingQueue<Session>();

    /**
     * Thread pool manager for the reader task
     */
    protected ThreadPoolManager threadPoolManager;

    /**
     * Task in charge of the message receiving via the {@link #messageCodecItf}
     */
    protected ReaderTask readerTask = new ReaderTask();

    /**
     * Map containing {@link ChannelFactory channel factories} identified by
     * their {@link IPExportIdentifier} (more precisely by their port because
     * the hostname is the same for each {@link ChannelFactory channel factory}
     * ).
     */
    protected Map<IPExportIdentifier, ChannelFactory> exportedChannels = new HashMap<IPExportIdentifier, ChannelFactory>();

    // ------------------------------------------------------------------------
    // ---
    // Attribute fields
    // ------------------------------------------------------------------------
    // ---

    /**
     * Name of the local host. It should be set to "localhost" in most cases
     * except when virtual IP are used
     */
    @Attribute(argument = "hostName")
    protected String hostName;

    /**
     * The listening port on which the TCP/IP protocol will wait incoming
     * connection for an exported channel.
     */
    @Attribute(argument = "port")
    protected int port;

    // Cannot set a timeout to the SSLSocketFactory!
    // @Attribute(argument = "connectionTimeout")
    // protected int connectionTimeout;

    /**
     * The number of connection retries before aborting.
     */
    @Attribute(argument = "connectionRetry")
    protected short connectionRetry;

    /**
     * Specify if the Nagle's algorithm is disabled.
     */
    @Attribute(argument = "tcpNoDelay")
    protected boolean tcpNoDelay;

    /**
     * the <code>SO_TIMEOUT</code> parameter of created sockets.
     */
    @Attribute(argument = "soTimeout")
    protected int soTimeout;

    /**
     * the <code>SO_LINGER</code> parameter of created sockets. A negative value
     * disable the linger on close.
     */
    @Attribute(argument = "soLinger")
    protected int soLinger;

    /**
     * the <code>SO_KEEPALIVE</code> parameter of created sockets. A {@code
     * false} value disable the keeping of an alive connection on the socket.
     */
    @Attribute(argument = "soKeepAlive")
    protected boolean soKeepAlive;

    /**
     * The keystore file
     */
    @Attribute(argument = "keyStoreFile")
    protected String keyStoreFile;

    /**
     * the keystore password
     */
    @Attribute(argument = "keyStorePassword")
    protected String keyStorePassword;

    /**
     * the key password
     */
    @Attribute(argument = "keyPassword")
    protected String keyPassword;

    /**
     * The truststore file
     */
    @Attribute(argument = "trustStoreFile")
    protected String trustStoreFile;

    /**
     * the truststore password
     */
    @Attribute(argument = "trustStorePassword")
    protected String trustStorePassword;

    // ------------------------------------------------------------------------
    // ---
    // Client interfaces
    // ------------------------------------------------------------------------
    // ---

    @Requires(name = ConnectionFactoryItf.ITF_NAME)
    protected ConnectionFactoryItf connectionFactoryItf;

    @Requires(name = "message-manager")
    protected MessageManagerType messageManagerItf;

    @Requires(name = "message-codec")
    protected MessageCodec messageCodecItf;

    @Requires(name = "task-manager")
    protected TaskManager taskManagerItf;

    // ------------------------------------------------------------------------
    // ---
    // SSL
    // ------------------------------------------------------------------------
    // ---

    /**
     * The key manager factory
     */
    protected KeyManagerFactory keyManagerFactory;

    /**
     * The keystore
     */
    protected KeyStore keyStore;

    /**
     * The truststore
     */
    protected KeyStore trustStore;

    /**
     * The trust manager factory
     */
    protected TrustManagerFactory trustManagerFactory;

    /**
     * The SSL context
     */
    protected SSLContext sslContext;

    /**
     * The SSL socket factory
     */
    protected SSLSocketFactory sslSocketFactory;

    /**
     * The SSL server socket factory
     */
    protected SSLServerSocketFactory sslServerSocketFactory;

    /**
     * Is SSL initialized ?
     */
    protected Boolean sslInitialized = new Boolean(false);

    // ------------------------------------------------------------------------
    // ---
    // Implementation of the ChannelProtocol interface
    // ------------------------------------------------------------------------
    // ---

    /**
     * @see ChannelProtocol#export(ChannelFactory, Map)
     */
    public ExportIdentifier export(ChannelFactory channel, Map<String, Object> hints)
            throws ExportException {
        Number portN = null;
        Number rangeN = null;
        // first non well-known port
        int port = 0;
        int range = 0;
        if (hints != null) {
            portN = (Number) hints.get(PORT);
            rangeN = (Number) hints.get(PORT_RANGE);
        }

        if (portN == null) {
            if (rangeN == null) {
                range = -1;
            } else {
                range = rangeN.intValue();
            }
        } else {
            port = portN.intValue();
            if (rangeN != null) {
                range = rangeN.intValue();
            }
        }
        return export(channel, port, range);
    }

    /**
     * @see ChannelProtocol#unexport(ExportIdentifier)
     */
    public void unexport(ExportIdentifier exportIdentifier) throws InvalidExportIdentifierException {
        if (!(exportIdentifier instanceof IPExportIdentifier)) {
            throw new InvalidExportIdentifierException(
                    "this protocol requires a IPExportIdentifier.", exportIdentifier);
        }
        connectionFactoryItf.removeServerSocket((IPExportIdentifier) exportIdentifier);
    }

    /**
     * @param hints
     *      if "localexportidentifier" key is set, this value is used for the
     *      (localaddr,localport) of the socket
     * @see ChannelProtocol#bind(ExportIdentifier, IncomingPush, Map)
     */
    synchronized public OutgoingPush bind(ExportIdentifier exportId, IncomingPush toClientPush,
            Map<String, Object> hints) throws InvalidExportIdentifierException, BindException {
        LoggerUtil.start(logger);
        
        // TODO check local binding.
        if (!(exportId instanceof IPExportIdentifier)) {
            throw new InvalidExportIdentifierException(
                    "this protocol requires a IPExportIdentifier.", exportId);
        }
        IPExportIdentifier identifier = (IPExportIdentifier) exportId;
        Socket socket;
        try {
            IPExportIdentifier localExportIdentifier = null;
            if (hints != null) {
                Object o = hints.get(LOCAL_EXPORT_IDENTIFIER_KEY);
                if (o instanceof IPExportIdentifier) {
                    logger.log(BasicLevel.ERROR, "localExportIdentifier specified : "
                            + localExportIdentifier);
                    localExportIdentifier = (IPExportIdentifier) o;
                } else {
                    logger.log(BasicLevel.ERROR, "the local export identifier specified in "
                            + "the hints map is not an instance of IPExportIdentifier : "
                            + o.getClass().getName());
                }
            }
            socket = createConnection(identifier, localExportIdentifier);
        } catch (IOException e) {
            throw new BindException("Unable to open connection to specified export : ", exportId, e);
        }   catch (IllegalStateException e) {
            throw new BindException("Unable to open connection to specified export : ", exportId, e);
        }

        Session session;
        try {
            session = new Session(socket, identifier);
        } catch (IOException e) {
            silentClose(socket);
            throw new BindException("unable to create session for opened socket : ", exportId, e);
        }

        // wait for acknowledgment message.
        waitAknowledgment(exportId, session, toClientPush);
        initSession(session, toClientPush);
        LoggerUtil.returnAfter(logger);
        return session;
    }

    private void waitAknowledgment(ExportIdentifier exportId, Session session,
            IncomingPush toClientPush) throws BindException {
        Message message;
        try {
            message = messageCodecItf.decode(session.codecInputOutput);
        } catch (IOException e) {
            silentClose(session);
            throw new BindException("Unable to receive acknowledgment message", exportId, e);
        }
        ExceptionChunk chunk = (ExceptionChunk) messageManagerItf.getChunk(message,
                BIND_REPLY_CHUNK_NAME);
        if (chunk == null) {
            if (messageManagerItf.getChunk(message, ChannelFactory.BIND_REPLY_CHUNK_NAME) != null) {
                try {
                    toClientPush.incomingPush(message);
                } catch (PushException e) {
                    Error.error("First received message on socket does contain "
                            + "a bind-reply chunk for the channel factory "
                            + "and it cannot be transmitted to it", logger);
                }
                // recursive call
                waitAknowledgment(exportId, session, toClientPush);
            } else {
                silentClose(session);
                Error.error("First received message on socket does not contain "
                        + "acknowledgment chunk", logger);
            }
        } else {
            if (chunk.getException() != null) {
                throw new BindException("Remote exception", exportId, chunk.getException());
            }
        }
        messageManagerItf.deleteMessage(message);
    }

    // ------------------------------------------------------------------------
    // ---
    // Implementation of the TCPIPProtocol interface
    // ------------------------------------------------------------------------
    // ---

    /**
     * @see org.objectweb.dream.protocol.Protocol#createExportIdentifier(Map,
     *  ExportIdentifier[])
     */
    public ExportIdentifier createExportIdentifier(Map<String, Object> info, ExportIdentifier[] next)
            throws InvalidExportIdentifierException {
        if (next != null && next.length != 0) {
            throw new InvalidExportIdentifierException("TCP/IP protocol is a leaf "
                    + "in the protocol graph, it cannot have next export identifier");
        }
        Object o;
        if (info == null || (o = info.get(ADDRESS)) == null) {
            throw new InvalidExportIdentifierException("Can't find address in info map");
        }
        String hostName;
        if (o instanceof String) {
            hostName = (String) o;
        } else if (o instanceof InetAddress) {
            hostName = ((InetAddress) o).getCanonicalHostName();
        } else {
            throw new InvalidExportIdentifierException(
                    "Invalid address object in info map, must be an InetAddress or a String");
        }

        o = info.get(PORT);
        if (o == null) {
            throw new InvalidExportIdentifierException("Can't find port in info map");

        }
        if (!(o instanceof Number)) {
            throw new InvalidExportIdentifierException(
                    "Invalid port object in info map, must be a Number");
        }
        int port = ((Number) o).intValue();
        return createExportIdentifier(hostName, port);
    }

    /**
     * @see TCPIPProtocol#export(ChannelFactory, int, int)
     */
    synchronized public IPExportIdentifier export(ChannelFactory channel, int port, int range)
            throws ExportException {
        LoggerUtil.start(logger);

        synchronized (sslInitialized) {
            if (!sslInitialized) {
                throw new ExportException("SSLProtocol is not initialized");
            }
        }

        synchronized (this) {
            if (port == 0) {
                port = this.port;
                // this.port;
            }
        }
        int maxPort;
        if (range < 0) {
            maxPort = 65536;
        } else {
            maxPort = port + range;
        }
        ServerSocket serverSocket = null;
        for (int p = port; p <= maxPort; p++) {
            try {
                serverSocket = sslServerSocketFactory.createServerSocket(p);
                break;
            } catch (Exception e) {
                e.printStackTrace();

                if (p == maxPort) {
                    throw new ExportException("Unable to open server socket : " + e.getMessage());
                }
            }
        }

        IPExportIdentifier exportIdentifier = new IPExportIdentifier(hostName, serverSocket
                .getLocalPort());
        exportedChannels.put(exportIdentifier, channel);
        connectionFactoryItf.addServerSocket(serverSocket, exportIdentifier);
        LoggerUtil.returnAfter(logger);
        return exportIdentifier;
    }

    /**
     * @see TCPIPProtocol#bind(String, int, IncomingPush)
     */
    public OutgoingPush bind(String hostName, int port, IncomingPush toClientPush)
            throws InvalidExportIdentifierException, BindException {
        return bind(createExportIdentifier(hostName, port), toClientPush, null);
    }

    /**
     * @see TCPIPProtocol#createExportIdentifier(String, int)
     */
    public IPExportIdentifier createExportIdentifier(String hostName, int port) {
        return new IPExportIdentifier(hostName, port);
    }

    // ------------------------------------------------------------------------
    // ---
    // Implementation of the ConnectionManager interface
    // ------------------------------------------------------------------------
    // ---

    /**
     * @see ConnectionManager#newConnection(Socket, IPExportIdentifier)
     */
    public void newConnection(Socket socket, IPExportIdentifier identifier) {
        LoggerUtil.start(logger);
        ChannelFactory channel = exportedChannels.get(identifier);
        if (DEFENSIVE_CHECK && channel == null) {
            Error.bug(logger);
        }

        Message message = messageManagerItf.createMessage();
        ExceptionChunk chunk = newExceptionChunk();
        messageManagerItf.addChunk(message, BIND_REPLY_CHUNK_NAME, chunk);
        Session session;
        try {
            session = new Session(socket, identifier);
        } catch (IOException e) {
            silentClose(socket);
            logger.log(BasicLevel.ERROR, "Unable to create session for new incoming socket", e);
            return;
        }
        IncomingPush incomingPush = null;
        try {
            incomingPush = channel.instantiate(session);
        } catch (IOException e) {
            chunk.setException(e);
            logger.log(BasicLevel.DEBUG, "Exception caught while instantiating channel factory", e);
        }
        // send the bind reply of the export/bind pattern
        try {
            session.outgoingPush(message);
        } catch (IOPushException e) {
            logger.log(BasicLevel.ERROR, "Unable to send reply to client, close socket", e);
            silentClose(session);
            return;
        }
        if (incomingPush != null) {
            initSession(session, incomingPush);
        }
        LoggerUtil.end(logger);
    }

    // ------------------------------------------------------------------------
    // ---
    // Utility method
    // ------------------------------------------------------------------------
    // ---

    private ExceptionChunk newExceptionChunk() {
        if (exceptionChunkFactory == null) {
            exceptionChunkFactory = messageManagerItf.getChunkFactory(ExceptionChunk.class);
        }
        return messageManagerItf.createChunk(exceptionChunkFactory);
    }

    // ------------------------------------------------------------------------
    // ---
    // Inner Class
    // ------------------------------------------------------------------------
    // ---

    protected class Session implements OutgoingPush {

        boolean closed = false;

        Socket socket;

        CodecInputOutput codecInputOutput = new SocketCodecInputOutput();

        IPExportIdentifier exportIdentifier;

        IncomingPush upperIncomingPushItf;

        Session(Socket socket, IPExportIdentifier exportIdentifier) throws IOException {
            this.socket = socket;
            this.exportIdentifier = exportIdentifier;
            codecInputOutput.setInput(socket.getInputStream());
            codecInputOutput.setOutput(socket.getOutputStream());
        }

        synchronized void closeSession() throws IOException {
            LoggerUtil.start(logger);
            closed = true;
            codecInputOutput.close();
            if (!socket.isClosed()) {
                socket.close();
            }
            LoggerUtil.end(logger);
        }

        // --------------------------------------------------------------------
        // -----
        // Implementation of the OutgoingPush interface
        // --------------------------------------------------------------------
        // -----

        /**
         * @see OutgoingPush#outgoingPush(Message)
         */
        public synchronized void outgoingPush(Message message) throws IOPushException {
            LoggerUtil.start(logger);
            if (closed) {
                throw new IOPushException("Session closed");
            }
            try {
                if (DEFENSIVE_CHECK && logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "Sending a message via " + this);
                messageCodecItf.encode(codecInputOutput, message);
            } catch (IOException e) {
                if (!closed) {
                    logger.log(BasicLevel.INFO,
                            "An error occurs while writing message on TCP/IP socket", e);
                    sessionError(this, e);
                }
                throw new IOPushException("unable to send message", e);
            }
            messageManagerItf.deleteMessage(message);
            LoggerUtil.end(logger);
        }

        /**
         * @see OutgoingPush#outgoingClose(IncomingPush)
         */
        public void outgoingClose(IncomingPush incomingPush) throws IOException {
            closeSession();
        }

        /**
         * Nice format for output
         */
        @Override
        public String toString() {
            String res = "";
            if (exportIdentifier == null) {
                res = "AnonymousTCPSession@" + Integer.toHexString(hashCode()) + "(";
            } else {
                res = "TCPSession@" + Integer.toHexString(hashCode()) + "(id:" + exportIdentifier
                        + ", localSocket(" + socket.getLocalAddress() + ":" + socket.getLocalPort()
                        + "), remoteSocket(" + socket.getInetAddress() + ":" + socket.getPort()
                        + "), ";
            }
            res = "State : " + (closed ? "closed" : "open") + ")";
            return res;
        }
    }

    protected class ReaderTask extends AbstractTask {

        public ReaderTask() {
            super("TCP/IP reader task");
        }

        // --------------------------------------------------------------------
        // -----
        // Implementation of the Task interface
        // --------------------------------------------------------------------
        // -----

        /**
         * @see org.objectweb.dream.control.activity.task.Task#execute(Object)
         */
        public Object execute(Object hints) throws InterruptedException {
            Session session;
            session = availableReadingSessions.poll(1, TimeUnit.SECONDS);
            if (session == null) {
                logger.log(BasicLevel.INFO, "No available session for the reader task");
                return STOP_EXECUTING;
            }
            if (session.closed) {
                logger.log(BasicLevel.INFO, session + " is closed, this task will be stopped");
                return STOP_EXECUTING;
            }
            Message message;
            try {
                message = messageCodecItf.decode(session.codecInputOutput);
            } catch (IOException e) {
                synchronized (this) {
                    // test if the session was properly closed and causes this
                    // exception
                    // (= closeSession() method called)
                    if (!session.closed) {
                        logger.log(BasicLevel.INFO,
                                "An error occurs while reading message on SSL socket", e);
                        sessionError(session, e);
                    }
                }
                return STOP_EXECUTING;
            }
            try {
                session.upperIncomingPushItf.incomingPush(message);
                logger.log(BasicLevel.DEBUG, "Sending a message to the upper level protocol");
            } catch (PushException e) {
                logger.log(BasicLevel.ERROR,
                        "Exception catch while pushing incoming message. delete message.", e);
                messageManagerItf.deleteMessage(message);
            }
            synchronized (sessions) {
                availableReadingSessions.put(session);
            }
            return EXECUTE_AGAIN;
        }

        /**
         * @see
         *  org.objectweb.dream.control.activity.task.Task#registered(Object)
         */
        public void registered(Object controlItf) {
            setThreadPoolManager((ThreadPoolManager) controlItf);
        }
    }

    // ------------------------------------------------------------------------
    // ---
    // Utility methods
    // ------------------------------------------------------------------------
    // ---

    /**
     * Close the session and ignore potential IOException
     * 
     * @param session
     *      the session to close
     */
    protected final void silentClose(Session session) {
        try {
            session.closeSession();
        } catch (IOException e) {
            // ignore
            logger.log(BasicLevel.DEBUG, "An error occured while closing " + session + " : "
                    + e.getMessage());
        }
    }

    /**
     * Close the socket and ignore potential IOException
     * 
     * @param socket
     *      the socket to close
     */
    protected final void silentClose(Socket socket) {
        LoggerUtil.start(logger);
        try {
            socket.shutdownInput();
            socket.shutdownOutput();
            socket.close();
            if (DEFENSIVE_CHECK && !socket.isClosed()) {
                Error.bug(logger);
            }
            if (DEFENSIVE_CHECK && !socket.isInputShutdown()) {
                Error.bug(logger);
            }
            if (DEFENSIVE_CHECK && !socket.isOutputShutdown()) {
                Error.bug(logger);
            }
            LoggerUtil.end(logger);
        } catch (IOException e) {
            // ignore
            logger.log(BasicLevel.DEBUG, "An error occured while closing a socket : "
                    + e.getMessage());
        }
    }

    /**
     * @param session
     * @param upperIncomingPush
     *      incoming interface of the upper session
     * @return
     */
    protected synchronized Session initSession(Session session, IncomingPush upperIncomingPush) {
        LoggerUtil.start(logger);
        session.upperIncomingPushItf = upperIncomingPush;
        synchronized (sessions) {
            sessions.add(session);
            if (session.upperIncomingPushItf != null) {
                try {
                    availableReadingSessions.put(session);
                } catch (InterruptedException e1) {
                    Error.bug(logger, e1);
                }
                if (threadPoolManager != null) {
                    try {
                        logger.log(BasicLevel.DEBUG, "add a thread in thread pool");
                        threadPoolManager.addThread(readerTask);
                    } catch (Exception e) {
                        Error.bug(logger, e);
                    }
                }
            }
            LoggerUtil.returnAfter(logger);
        }
        return session;
    }

    protected void setThreadPoolManager(ThreadPoolManager threadPoolManager) {
        this.threadPoolManager = threadPoolManager;
        synchronized (sessions) {
            for (Session s : sessions) {
                try {
                    logger.log(BasicLevel.DEBUG, "add a thread in thread pool for " + s);
                    threadPoolManager.addThread(readerTask);
                    availableReadingSessions.put(s);
                } catch (ThreadPoolOverflowException e) {
                    logger.log(BasicLevel.WARN, "Unable to add reader thread", e);
                } catch (IllegalTaskException e) {
                    Error.bug(logger, e);
                } catch (InterruptedException e) {
                    Error.bug(logger, e);
                }
            }
        }
    }

    protected void sessionError(Session session, Exception exception) {
        synchronized (sessions) {
            sessions.remove(session);
        }
        silentClose(session);
        if (session.upperIncomingPushItf != null) {
            // session is a server session, inform channel.
            session.upperIncomingPushItf.incomingClosed(session, exception);
        }
    }

    /**
     * Try to create a connection toward the destination specified via the
     * export identifier. the number of try depends on the value of the
     * <code>connectionRetry</code> fractal attribute.
     * 
     * @param exportIdentifier
     *      identifier the destination
     * @param localExportIdentifier
     * @return the created socket
     * @throws IOException
     *      if an error occured while connecting
     * @throws CertificateException
     * @throws NoSuchAlgorithmException
     * @throws KeyStoreException
     * @throws UnrecoverableKeyException
     * @throws KeyManagementException
     */
    protected Socket createConnection(IPExportIdentifier exportIdentifier,
            IPExportIdentifier localExportIdentifier) throws IOException {
        IOException exception = null;
        Socket clientSocket = null;

        synchronized (sslInitialized) {
            if (!sslInitialized) {
                throw new IllegalStateException("SSLProtocol is not initialized");
            }
        }

        for (int i = 0; i <= connectionRetry; i++) {
            try {
                logger.log(BasicLevel.DEBUG, "try to open a SSL connection");

                if (localExportIdentifier == null) {
                    clientSocket = sslSocketFactory.createSocket(exportIdentifier.getHostName(),
                            exportIdentifier.getPort());
                } else {
                    clientSocket = sslSocketFactory.createSocket(exportIdentifier.getHostName(),
                            exportIdentifier.getPort(), InetAddress.getByName(localExportIdentifier
                                    .getHostName()), localExportIdentifier.getPort());
                }
                logger.log(BasicLevel.DEBUG, "connected to " + exportIdentifier.getHostName() + ":"
                        + exportIdentifier.getPort());

                // TODO: Check if the setting of the parameters is working after
                // having connected the SSL socket
                setSocketOption(clientSocket);

                break;
            } catch (IOException exc) {
                exception = exc;
                logger.log(BasicLevel.INFO, "connection failed on address "
                        + exportIdentifier.getHostName() + ":" + exportIdentifier.getPort()
                        + ". Another attempt will be done at " + i * 250 + " ms");
                try {
                    Thread.sleep(i * 250);
                } catch (InterruptedException ignored) {
                    break;
                }
            }
        }

        if (exception != null) {
            throw exception;
        } else {
            return clientSocket;
        }
    }

    protected void setSocketOption(Socket sock) throws IOException {
        sock.setTcpNoDelay(tcpNoDelay);
        sock.setSoTimeout(soTimeout);
        sock.setKeepAlive(soKeepAlive);
        if (soLinger >= 0) {
            sock.setSoLinger(true, soLinger);
        } else {
            sock.setSoLinger(false, 0);
        }
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Options for socket (" + sock.getInetAddress() + ","
                    + sock.getPort() + ") : tcpNoDelay = " + sock.getTcpNoDelay() + " soLinger = "
                    + sock.getSoLinger() + " soTimeout = " + sock.getSoTimeout()
                    + " soKeepAlive = " + sock.getKeepAlive() + " sendBufferSize = "
                    + sock.getSendBufferSize());
        }
    }

    // -------------------------------------------------------------------------
    // Dream Components Lifecycle Methods
    // -------------------------------------------------------------------------

    /**
     * Method executed during the first start of the component
     */
    @DreamLifeCycle(on = DreamLifeCycleType.FIRST_START)
    synchronized protected void beforeFirstStart(Component componentItf)
            throws IllegalLifeCycleException {

        logger.log(BasicLevel.DEBUG,
                "First start of the component : SSl initialisation + SSL reader task added");

        try {
            synchronized (this.sslInitialized) {
                if (!sslInitialized) {
                    initSSL();
                }
            }

            Map<String, Object> hints = new HashMap<String, Object>();
            hints.put("thread", "pool");
            Util.addTask(componentItf, readerTask, hints);
        } catch (Exception e) {
            logger.log(BasicLevel.FATAL, e);
            throw new IllegalLifeCycleException("Failed to Prepare first start. Caused : "
                    + e.getClass().getSimpleName() + " - " + e.getMessage());
        }
    }

    /**
     * @see
     *  org.objectweb.dream.control.lifecycle.PrepareStopLifeCycleController#
     *  prepareStopFc()
     */
    @DreamLifeCycle(on = DreamLifeCycleType.PREPARE_TO_STOP)
    public void prepareStopFc() throws IllegalLifeCycleException {
        synchronized (sessions) {

            logger.log(BasicLevel.DEBUG,
                    "Prepare stop of the component : Close every opened SSL sockets");

            for (Session s : sessions) {
                logger.log(BasicLevel.DEBUG, "Going to close " + s);
                silentClose(s);
            }
            sessions.clear();
        }
    }

    protected synchronized void initSSL() throws KeyStoreException, NoSuchAlgorithmException,
            CertificateException, FileNotFoundException, IOException, KeyManagementException,
            UnrecoverableKeyException {
        keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());

        keyStore = KeyStore.getInstance(KeyStore.getDefaultType());
        keyStore.load(new FileInputStream(keyStoreFile), keyStorePassword.toCharArray());

        trustStore = KeyStore.getInstance(KeyStore.getDefaultType());
        trustStore.load(new FileInputStream(trustStoreFile), trustStorePassword.toCharArray());

        keyManagerFactory.init(keyStore, keyPassword.toCharArray());

        trustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory
                .getDefaultAlgorithm());

        trustManagerFactory.init(trustStore);

        sslContext = SSLContext.getInstance("TLS");
        sslContext.init(keyManagerFactory.getKeyManagers(), trustManagerFactory.getTrustManagers(),
                null);

        sslSocketFactory = sslContext.getSocketFactory();

        sslServerSocketFactory = sslContext.getServerSocketFactory();

        sslInitialized = true;
    }

    // -------------------------------------------------------------------------
    // Attribute Controller interface
    // -------------------------------------------------------------------------

    /**
     * @see TCPIPProtocolImplAttributeController#setHostName(String)
     */
    public void setHostName(String hostname) throws UnknownHostException {
        if (hostname.trim().equalsIgnoreCase("localhost")) {
            this.hostName = InetAddress.getLocalHost().getHostName();
        } else {
            // this call make a check on the given parameter
            this.hostName = InetAddress.getByName(hostname).getHostName();
        }
    }
}
