/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s):
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.channel.http;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;

import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.codec.CodecInputOutput;
import org.objectweb.dream.message.codec.MessageCodec;
import org.objectweb.dream.serializator.DeSerializatorItf;
import org.objectweb.dream.serializator.SerializatorItf;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * @author pgarcia
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
@Provides(interfaces={@Interface(name="message-codec", signature = MessageCodec.class)})
public class MessageCodecHelperHttpObjectStream implements MessageCodec
{
  /**
   * 
   */
  @Service
  Component                 ref;

  /**
   * 
   */
  @DreamMonolog
  Logger                    logger;

  /**
   * 
   */
  @Requires(name = "serializator")
  SerializatorItf<String>   xmlSerializator;

  /**
   * 
   */
  @Requires(name = "deserializator")
  DeSerializatorItf<String> xmlDeSerializator;

  // ---------------------------------------------------------------------------
  // Attribute field
  // ---------------------------------------------------------------------------

  /**
   * Attribute used to know if the created ObjectOuputStream/ObjectInputStream
   * must be reused of re-created for each encoding/decoding
   */
  @Attribute(argument = "reuseObjectStream")
  boolean                   reuseObjectStream;

  private final String      DEFAULT_HTTP_LINE_SEPARATOR            = "\r\n";

  /**
   * 
   */
  private final String      DEFAULT_HTTP_HEADER_REQUEST            = "POST /uri HTTP/1.1"
                                                                       + DEFAULT_HTTP_LINE_SEPARATOR;

  /**
   * 
   */
  private final String      DEFAULT_HTTP_HEADER_TARGET_HOST        = "Host: fakehost"
                                                                       + DEFAULT_HTTP_LINE_SEPARATOR;

  /**
   * 
   */
  private final String      DEFAULT_HTTP_HEADER_CONTENT_LENGTH_TAG = "Content-Length: ";

  /**
   * @see org.objectweb.dream.message.codec.MessageCodec#decode(org.objectweb.dream.message.codec.CodecInputOutput)
   */
  public Message decode(CodecInputOutput cio) throws IOException
  {
    Object input = cio.getInput();
    ObjectInputStream ois;
    if (input instanceof ObjectInputStream)
    {
      ois = (ObjectInputStream) input;
    }
    else
    {
      if (!(input instanceof InputStream))
      {
        throw new IOException("Unknown input : " + input);
      }
      ois = new ObjectInputStream((InputStream) input);
      cio.setInput(ois);
    }

    // message decoding
    String httpMessage = ois.readUTF();

    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "\nHTTP decoded request : \n" + httpMessage);
    }

    String xmlMessage = parse(httpMessage);

    if (!reuseObjectStream)
    {
      cio.setInput(input);
    }
    return xmlDeSerializator.deserialize(xmlMessage);
  }

  private String parse(String httpMessage)
  {
    String[] splittedMessage = httpMessage.split(DEFAULT_HTTP_LINE_SEPARATOR
        + DEFAULT_HTTP_LINE_SEPARATOR);
    if(! (splittedMessage.length ==2))
      Error.error("unvalid message format " + httpMessage, logger);
    return splittedMessage[1];
  }

  /**
   * @see org.objectweb.dream.message.codec.MessageCodec#encode(org.objectweb.dream.message.codec.CodecInputOutput,
   *      org.objectweb.dream.message.Message)
   */
  public void encode(CodecInputOutput cio, Message message) throws IOException
  {
    // ObjectOuputStream management
    Object output = cio.getOutput();
    ObjectOutputStream oos;
    if (output instanceof ObjectOutputStream)
    {
      oos = (ObjectOutputStream) output;
    }
    else
    {
      if (!(output instanceof OutputStream))
      {
        throw new IOException("Unknown output : " + output);
      }
      oos = new ObjectOutputStream((OutputStream) output);
      cio.setOutput(oos);
    }

    // Message encoding
    String httpMessage = xmlSerializator.serialize(message);
    // Add http header
    httpMessage = DEFAULT_HTTP_HEADER_REQUEST + DEFAULT_HTTP_HEADER_TARGET_HOST
        + DEFAULT_HTTP_HEADER_CONTENT_LENGTH_TAG + httpMessage.length()
        + DEFAULT_HTTP_LINE_SEPARATOR + DEFAULT_HTTP_LINE_SEPARATOR
        + httpMessage;
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger
          .log(BasicLevel.DEBUG, "\nHTTP received request : \n" + httpMessage);
    }
    oos.writeUTF(httpMessage);
    oos.flush();

    // Test if the ObjectOutput Stream must be reused or not
    if (!reuseObjectStream)
    {
      cio.setOutput(output);
    }
  }
}
