/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.serializator.ByteArrayChunk;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Fragmentation protocol. Fragments outgoing messages, and reassemble incoming
 * fragmented messages. Outgoing messages should contain only a {@link
 * ByteArrayChunk}, additional chunks will not be present in the reassembled
 * message.
 */
@DreamComponent(controllerDesc = "activeDreamUnstoppablePrimitive")
@Provides(interfaces = { @Interface(name = MessagePassingProtocol.ITF_NAME, signature = MessagePassingProtocol.class) })
public class FragmentationProtocolImpl implements MessagePassingProtocol {
    private ChunkFactoryReference<ByteArrayChunk> byteArrayChunkFactory;

    private ChunkFactoryReference<FragmentationChunk> fragmentationChunkFactory;

    // ------------------------------------------------------------------------
    // ---
    // Attribute fields
    // ------------------------------------------------------------------------
    // ---

    /**
     * the <code>fragmentationSize</code> attribute value to set.
     */
    @Attribute(argument = "fragmentationSize")
    protected int fragmentationSize;

    // ------------------------------------------------------------------------
    // ---
    // Client interfaces
    // ------------------------------------------------------------------------
    // ---
    @Requires(name = "lower-protocol")
    protected MessagePassingProtocol lowerLevelProtocolItf;

    @Requires(name = "message-manager")
    protected MessageManagerType messageManagerItf;

    // ------------------------------------------------------------------------
    // --
    // Services interfaces
    // ------------------------------------------------------------------------
    // --

    /**
     * Component reference
     */
    @Service
    Component weaveableC;

    /**
     * Logger of the component
     */
    @DreamMonolog()
    protected Logger logger;

    // ------------------------------------------------------------------------
    // ---
    // Implementation of the MessagePassingProtocol interface
    // ------------------------------------------------------------------------
    // ---

    /**
     * @see MessagePassingProtocol#export(IncomingPush, Map)
     */
    public MessagePassingOutgoingPush export(IncomingPush incomingPushItf, Map<String, Object> hints)
            throws ExportException {
        Session session = new Session(incomingPushItf);
        session.initialize(lowerLevelProtocolItf.export(session, hints));
        return session;
    }

    /**
     * @see org.objectweb.dream.protocol.Protocol#createExportIdentifier(Map,
     *  ExportIdentifier[])
     */
    public ExportIdentifier createExportIdentifier(Map<String, Object> info, ExportIdentifier[] next)
            throws InvalidExportIdentifierException {
        return lowerLevelProtocolItf.createExportIdentifier(info, next);
    }

    // ------------------------------------------------------------------------
    // ---
    // Utility method
    // ------------------------------------------------------------------------
    // ---

    private ByteArrayChunk newByteArrayChunk() {
        if (byteArrayChunkFactory == null) {
            byteArrayChunkFactory = messageManagerItf.getChunkFactory(ByteArrayChunk.class);
        }
        return messageManagerItf.createChunk(byteArrayChunkFactory);
    }

    private FragmentationChunk newFragmentationChunk() {
        if (fragmentationChunkFactory == null) {
            fragmentationChunkFactory = messageManagerItf.getChunkFactory(FragmentationChunk.class);
        }
        return messageManagerItf.createChunk(fragmentationChunkFactory);
    }

    // ------------------------------------------------------------------------
    // ---
    // Inner Class
    // ------------------------------------------------------------------------
    // ---

    protected class Session implements MessagePassingOutgoingPush, IncomingPush {
        private final IncomingPush upperIncomingPush;

        private MessagePassingOutgoingPush lowerOutgoingPush;

        private int currentId = 0;

        private Map<ExportIdentifier, FragmentationTable> fragmentMap = new HashMap<ExportIdentifier, FragmentationTable>();

        protected Session(IncomingPush upperIncomingPush) {
            this.upperIncomingPush = upperIncomingPush;
        }

        protected void initialize(MessagePassingOutgoingPush lowerPush) {
            this.lowerOutgoingPush = lowerPush;
        }

        // --------------------------------------------------------------------
        // -----
        // Implementation of the MessagePassingOutgoingPush interface
        // --------------------------------------------------------------------
        // -----

        /**
         * @see MessagePassingOutgoingPush#getLocalExportIdentifier()
         */
        public ExportIdentifier getLocalExportIdentifier() {
            return lowerOutgoingPush.getLocalExportIdentifier();
        }

        public void outgoingPush(Message message, ExportIdentifier to)
                throws InvalidExportIdentifierException, IOPushException {
            ByteArrayChunk byteArrayChunk = (ByteArrayChunk) messageManagerItf.getChunk(message,
                    ByteArrayChunk.DEFAULT_NAME);
            if (byteArrayChunk == null) {
                throw new IOPushException("Cannot find byte array chunk in outgoing message.");
            }

            byte[] byteArray = byteArrayChunk.getByteArray();
            long size = byteArray.length;

            if (size > fragmentationSize) {
                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "message size is " + size
                            + ", will fragment (frag_size == " + fragmentationSize + ')');
                }
                byte[] fragments[];

                int fragmentationId = currentId++;
                int numFrags = 0;

                fragments = fragmentBuffer(byteArray);
                numFrags = fragments.length;

                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "fragmenting packet (size=" + byteArray.length
                            + ") into " + numFrags + " fragment(s) [frag_size=" + fragmentationSize
                            + ']');
                }

                for (int fragId = 0; fragId < numFrags; fragId++) {
                    Message fragMessage = messageManagerItf.createMessage();
                    // Add the fragment (i.e. the ByteArrayChunk)
                    ByteArrayChunk byteArrayChunkFrag = newByteArrayChunk();
                    byteArrayChunkFrag.setByteArray(fragments[fragId]);
                    messageManagerItf.addChunk(fragMessage, ByteArrayChunk.DEFAULT_NAME,
                            byteArrayChunkFrag);

                    // Add the FragmentationChunk
                    FragmentationChunk fragmentationChunk = newFragmentationChunk();
                    fragmentationChunk.setFragmentationId(fragmentationId);
                    fragmentationChunk.setFragmentId(fragId);
                    fragmentationChunk.setFramentFrom(lowerOutgoingPush.getLocalExportIdentifier());
                    fragmentationChunk.setNumFrags(numFrags);
                    messageManagerItf.addChunk(fragMessage, FragmentationChunk.DEFAULT_NAME,
                            fragmentationChunk);
                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG, "fragment's fragmentation chunk is "
                                + fragmentationChunk);
                    }
                    try {
                        lowerOutgoingPush.outgoingPush(fragMessage, to);
                    } catch (IOPushException e) {
                        messageManagerItf.deleteMessage(fragMessage);
                        throw e;
                    } catch (InvalidExportIdentifierException e) {
                        messageManagerItf.deleteMessage(fragMessage);
                        throw e;
                    }
                }
                messageManagerItf.deleteMessage(message);
            } else {
                messageManagerItf.addChunk(message, ByteArrayChunk.DEFAULT_NAME, byteArrayChunk);
                lowerOutgoingPush.outgoingPush(message, to);
            }
        }

        public void outgoingClose(IncomingPush incomingPush) throws IOException {
            lowerOutgoingPush.outgoingClose(this);
        }

        // --------------------------------------------------------------------
        // -----
        // Implementation of the IncomingPush interface
        // --------------------------------------------------------------------
        // -----

        /**
         * @see IncomingPush#incomingPush(Message)
         */
        public void incomingPush(Message message) throws PushException {
            FragmentationChunk fragmentationChunk = (FragmentationChunk) messageManagerItf
                    .getChunk(message, FragmentationChunk.DEFAULT_NAME);

            if (fragmentationChunk != null) {
                messageManagerItf.removeChunk(message, FragmentationChunk.DEFAULT_NAME);
                FragmentationTable fragTable = null;
                ExportIdentifier source = fragmentationChunk.getFramentFrom();
                byte[] m;

                if (logger.isLoggable(BasicLevel.DEBUG)) {
                    logger.log(BasicLevel.DEBUG, "Received message, FragmentationChunk is "
                            + fragmentationChunk);
                }
                synchronized (fragmentMap) {
                    fragTable = fragmentMap.get(source);
                    if (fragTable == null) {
                        fragTable = new FragmentationTable();
                        fragmentMap.put(source, fragTable);
                    }
                }
                ByteArrayChunk byteArrayChunk = (ByteArrayChunk) messageManagerItf.getChunk(
                        message, ByteArrayChunk.DEFAULT_NAME);
                m = fragTable.add(fragmentationChunk.getFragmentationId(), fragmentationChunk
                        .getFragmentId(), fragmentationChunk.getNumFrags(), byteArrayChunk
                        .getByteArray());
                if (m != null) {
                    byteArrayChunk.setByteArray(m);
                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG, "Assembled message is " + message);
                    }
                    upperIncomingPush.incomingPush(message);
                } else {
                    messageManagerItf.deleteMessage(message);
                }
            } else {
                upperIncomingPush.incomingPush(message);
            }

        }

        /**
         * @see IncomingPush#incomingClosed(Object, Exception)
         */
        public synchronized void incomingClosed(Object outgoingPush, Exception exception) {
            // never called on message passing protocol
            Error.bug(logger);
        }

    }

    private static class FragmentationTable {
        private Map<Integer, Entry> h = new HashMap<Integer, Entry>();

        private static class Entry {
            // the total number of fragment in this message
            int totFrags = 0;

            // each fragment is a byte buffer
            byte[] fragments[] = null;

            // the number of fragments we have received
            int numberOfFragsRecvd = 0;

            // the message ID
            long msgId = -1;

            private Entry(long msgId, int totFrags) {
                this.msgId = msgId;
                this.totFrags = totFrags;
                fragments = new byte[totFrags][];
            }

            private void set(int fragId, byte[] frag) {
                fragments[fragId] = frag;
                numberOfFragsRecvd++;
            }

            private boolean isComplete() {
                // first make the simple check
                if (numberOfFragsRecvd < totFrags) {
                    return false;
                }
                // then double check just in case
                for (int i = 0; i < fragments.length; i++) {
                    if (fragments[i] == null)
                        return false;
                }
                // all fragmentations have been received
                return true;
            }

            private byte[] assembleBuffer() {
                return defragmentBuffer(fragments);
            }

        }

        private synchronized byte[] add(int id, int fragId, int totFrags, byte[] fragment) {

            // initialize the return value to default not complete
            byte[] retval = null;

            Entry e = h.get(id);

            if (e == null) { // Create new entry if not yet present
                e = new Entry(id, totFrags);
                h.put(id, e);
            }

            e.set(fragId, fragment);
            if (e.isComplete()) {
                retval = e.assembleBuffer();
                h.remove(new Long(id));
            }

            return retval;
        }
    }

    // -------------------------------------------------------------------------
    // Utility methods
    // -------------------------------------------------------------------------

    private byte[][] fragmentBuffer(byte[] buf) {
        byte[] retval[];
        long totalSize = buf.length;
        int accumulatedSize = 0;
        byte[] fragment;
        int tmpSize = 0;
        int numFrags;
        int index = 0;

        numFrags = buf.length % fragmentationSize == 0 ? buf.length / fragmentationSize
                : buf.length / fragmentationSize + 1;
        retval = new byte[numFrags][];

        while (accumulatedSize < totalSize) {
            if (accumulatedSize + fragmentationSize <= totalSize) {
                tmpSize = fragmentationSize;
            } else {
                tmpSize = (int) (totalSize - accumulatedSize);
            }
            fragment = new byte[tmpSize];
            System.arraycopy(buf, accumulatedSize, fragment, 0, tmpSize);
            retval[index++] = fragment;
            accumulatedSize += tmpSize;
        }
        return retval;
    }

    private static byte[] defragmentBuffer(byte[] fragments[]) {
        int totalLength = 0;
        byte[] ret;
        int index = 0;

        if (fragments == null) {
            return null;
        }
        for (int i = 0; i < fragments.length; i++) {
            if (fragments[i] == null) {
                continue;
            }
            totalLength += fragments[i].length;
        }
        ret = new byte[totalLength];
        for (int i = 0; i < fragments.length; i++) {
            if (fragments[i] == null) {
                continue;
            }
            System.arraycopy(fragments[i], 0, ret, index, fragments[i].length);
            index += fragments[i].length;
        }
        return ret;
    }
}
