/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.control.activity.Util;
import org.objectweb.dream.control.activity.manager.TaskManager;
import org.objectweb.dream.control.activity.manager.ThreadPoolManager;
import org.objectweb.dream.control.activity.task.AbstractTask;
import org.objectweb.dream.control.activity.task.IllegalTaskException;
import org.objectweb.dream.control.activity.task.thread.ThreadPoolOverflowException;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.dream.dreamannotation.DreamLifeCycle;
import org.objectweb.dream.dreamannotation.DreamMonolog;
import org.objectweb.dream.dreamannotation.util.DreamLifeCycleType;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IPExportIdentifier;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.serializator.DeSerializatorItf;
import org.objectweb.dream.serializator.SerializatorItf;
import org.objectweb.dream.util.Error;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.util.AbstractComponent;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Implementation of the UDP protocol.
 */
@DreamComponent(controllerDesc = "activeDreamUnstoppablePrimitive")
public class UDPProtocolImpl implements UDPProtocolItf {

    private static final boolean DEFENSIVE_CHECK = true;

    protected static String hostName;

    // ------------------------------------------------------------------------
    // --
    // Services interfaces
    // ------------------------------------------------------------------------
    // --

    /**
     * Component reference
     */
    @Service
    Component weaveableC;

    /**
     * Logger of the component
     */
    @DreamMonolog()
    protected Logger logger;

    static {
        try {
            hostName = InetAddress.getLocalHost().getHostName();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private Set<Session> sessions = new HashSet<Session>();

    private Session firstAvailable = null;

    private Session lastAvailable = null;

    private ThreadPoolManager threadPoolManager;

    private ReaderTask readerTask = new ReaderTask();

    private ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;

    // ------------------------------------------------------------------------
    // ---
    // Attribute fields
    // ------------------------------------------------------------------------
    // ---

    /**
     * The <code>port</code> attribute's value.C
     */
    @Attribute(argument = "port")
    int port;

    /**
     * The size of buffer used in {@link java.net.DatagramPacket} when receiving
     * message.
     */
    @Attribute(argument = "recvBufferSize")
    int recvBufferSize;

    // ------------------------------------------------------------------------
    // ---
    // Client interfaces
    // ------------------------------------------------------------------------
    // ---

    @Requires(name = "message-manager")
    private MessageManagerType messageManagerItf;

    @Requires(name = SerializatorItf.ITF_NAME)
    private SerializatorItf<byte[]> serializatorItf;

    @Requires(name = DeSerializatorItf.ITF_NAME)
    private DeSerializatorItf<byte[]> deSerializatorItf;

    @SuppressWarnings("unused")
    @Requires(name = "task-manager")
    private TaskManager taskManagerItf;

    // ------------------------------------------------------------------------
    // ---
    // Implementation of the UDPProtocolItf interface
    // ------------------------------------------------------------------------
    // ---

    /**
     * @see MessagePassingProtocol#export(IncomingPush, Map)
     */
    public MessagePassingOutgoingPush export(IncomingPush incomingPushItf, Map<String, Object> hints)
            throws ExportException {
        Number portN = null;
        Number rangeN = null;
        String fromChunkName = null;
        int port = 0;
        int range = 0;
        if (hints != null) {
            Object o = hints.get(PORT);
            if (o != null && !(o instanceof Number)) {
                throw new ExportException("Invalid 'port' hint value, must be a Number");
            }
            portN = (Number) o;

            o = hints.get(PORT_RANGE);
            if (o != null && !(o instanceof Number)) {
                throw new ExportException("Invalid 'range' hint value, must be a Number");
            }
            rangeN = (Number) o;

            o = hints.get(FROM_CHUNK_NAME);
            if (o != null && !(o instanceof String)) {
                throw new ExportException("Invalid 'from-chunk-name' hint value, must be a String");
            }
            fromChunkName = (String) o;
        }

        if (portN != null) {
            port = portN.intValue();
            if (rangeN != null) {
                range = rangeN.intValue();
            }
        } else {
            synchronized (this) {
                port = this.port;
                if (this.port != 0) {
                    this.port++;
                }
            }
        }
        return export(incomingPushItf, port, range, fromChunkName);
    }

    /**
     * @see org.objectweb.dream.protocol.Protocol#createExportIdentifier(Map,
     *  ExportIdentifier[])
     */
    public ExportIdentifier createExportIdentifier(Map<String, Object> info, ExportIdentifier[] next)
            throws InvalidExportIdentifierException {
        if (next != null && next.length != 0) {
            throw new InvalidExportIdentifierException("UDP protocol is a leaf "
                    + "in the protocol graph, it cannot have next export identifier");
        }
        Object o;
        if (info == null) {
            throw new InvalidExportIdentifierException("Can't find address in info map");
        }
        o = info.get(ADDRESS);
        if (o == null) {
            throw new InvalidExportIdentifierException("Can't find address in info map");
        }
        String hostName;
        if (o instanceof String) {
            hostName = (String) o;
        } else if (o instanceof InetAddress) {
            hostName = ((InetAddress) o).getCanonicalHostName();
        } else {
            throw new InvalidExportIdentifierException(
                    "Invalid address object in info map, must be an InetAddress or" + " a String");
        }

        o = info.get(PORT);
        if (o == null) {
            throw new InvalidExportIdentifierException("Can't find port in info map");

        }
        if (!(o instanceof Number)) {
            throw new InvalidExportIdentifierException(
                    "Invalid port object in info map, must be a Number");
        }
        int port = ((Number) o).intValue();
        return createExportIdentifier(hostName, port);
    }

    /**
     * @see UDPProtocolItf#export(IncomingPush, int, int, String)
     */
    public MessagePassingOutgoingPush export(IncomingPush incomingPushItf, int port, int range,
            String fromChunkChunk) throws ExportException {
        Exception exception = null;
        for (int p = port; p <= port + range; p++) {
            Session session;
            try {
                session = new Session(incomingPushItf, p, fromChunkChunk);
            } catch (SocketException e) {
                exception = e;
                continue;
            }
            initSession(session);
            return session;
        }
        throw new ExportException("Unable to open datagram socket", exception);
    }

    /**
     * @see UDPProtocolItf#createExportIdentifier(String, int)
     */
    public IPExportIdentifier createExportIdentifier(String hostName, int port) {
        return new IPExportIdentifier(hostName, port);
    }

    // ------------------------------------------------------------------------
    // ---
    // Utility methods
    // ------------------------------------------------------------------------
    // ---

    private ExportIdentifierChunk newExportIdentifierChunk() {
        if (exportIdChunkFactory == null) {
            exportIdChunkFactory = messageManagerItf.getChunkFactory(ExportIdentifierChunk.class);
        }
        return messageManagerItf.createChunk(exportIdChunkFactory);
    }

    // ------------------------------------------------------------------------
    // ---
    // Inner Class
    // ------------------------------------------------------------------------
    // ---

    protected class Session implements MessagePassingOutgoingPush {
        Session nextAvailable;

        byte[] recvBuffer = new byte[recvBufferSize];

        boolean closed = false;

        IncomingPush upperIncomingPushItf;

        DatagramSocket socket;

        IPExportIdentifier localExportId;

        String fromChunkName;

        Session(IncomingPush itf, int port, String fromChunkName) throws SocketException {
            logger.log(BasicLevel.DEBUG, "initialization of UDP : ");
            upperIncomingPushItf = itf;
            this.fromChunkName = fromChunkName;
            if (port > 0) {
                socket = new DatagramSocket(port);
            } else {
                socket = new DatagramSocket();
            }
            localExportId = new IPExportIdentifier(hostName, socket.getLocalPort());
        }

        void receiveMessage() {
            DatagramPacket packet = new DatagramPacket(recvBuffer, recvBuffer.length);
            Message message;
            try {
                socket.receive(packet);
                if (packet != null) {
                    logger.log(BasicLevel.DEBUG, "packet from destination : ");
                    if (packet.getAddress().getCanonicalHostName() != null) {
                        logger.log(BasicLevel.INFO, packet.getAddress().getCanonicalHostName());
                    } else {
                        logger.log(BasicLevel.INFO, "address null ! ");
                    }
                    if (packet.getPort() != 0) {
                        logger.log(BasicLevel.INFO, ":" + packet.getPort());
                    } else {
                        logger.log(BasicLevel.INFO, "Port null ! ");
                    }
                } else {
                    System.out.println("packet null !");
                }
                message = deSerializatorItf.deserialize(recvBuffer);
            } catch (IOException e) {
                // test if the session was properly closed and causes this
                // exception
                // (= closeSession() method called)
                if (!this.closed) {
                    logger.log(BasicLevel.INFO,
                            "An error occurs while reading message on UDP socket", e);
                    sessionError(this, e);
                }
                return;
            }

            if (fromChunkName != null) {
                ExportIdentifierChunk chunk = newExportIdentifierChunk();
                chunk.setExportIdentifier(new IPExportIdentifier(packet.getAddress()
                        .getCanonicalHostName(), packet.getPort()));
                messageManagerItf.addChunk(message, fromChunkName, chunk);
            }
            try {
                upperIncomingPushItf.incomingPush(message);
            } catch (PushException e) {
                logger.log(BasicLevel.ERROR,
                        "Exception catch while pushing incoming message. delete message.", e);
                messageManagerItf.deleteMessage(message);
            }

        }

        // --------------------------------------------------------------------
        // -----
        // Implementation of the MessagePassingOutgoingPush interface
        // --------------------------------------------------------------------
        // -----

        /**
         * @see MessagePassingOutgoingPush#getLocalExportIdentifier()
         */
        public ExportIdentifier getLocalExportIdentifier() {
            return localExportId;
        }

        /**
         * @see MessagePassingOutgoingPush#outgoingPush(Message,
         *  ExportIdentifier)
         */
        public void outgoingPush(Message message, ExportIdentifier to)
                throws InvalidExportIdentifierException, IOPushException {
            synchronized (this) {
                if (closed) {
                    throw new IOPushException("Session closed");
                }
            }
            if (!(to instanceof IPExportIdentifier)) {
                throw new InvalidExportIdentifierException("Not a IPExportIdentifier", to);
            }
            IPExportIdentifier ipTo = (IPExportIdentifier) to;
            byte[] msg;
            try {
                msg = serializatorItf.serialize(message);
                DatagramPacket packet = new DatagramPacket(msg, msg.length, InetAddress
                        .getByName(ipTo.getHostName()), ipTo.getPort());
                socket.send(packet);
                logger.log(BasicLevel.DEBUG, "packet to destination : " + to);
                messageManagerItf.deleteMessage(message);
            } catch (IOException e) {
                throw new IOPushException(e);
            }
        }

        /**
         * @see MessagePassingOutgoingPush#outgoingClose(IncomingPush)
         */
        public synchronized void outgoingClose(IncomingPush incomingPush) throws IOException {
            closed = true;
            socket.close();
        }
    }

    protected class ReaderTask extends AbstractTask {

        public ReaderTask() {
            super("UDP reader task");
        }

        // --------------------------------------------------------------------
        // -----
        // Implementation of the Task interface
        // --------------------------------------------------------------------
        // -----

        /**
         * @see org.objectweb.dream.control.activity.task.Task#execute(Object)
         */
        public Object execute(Object hints) throws InterruptedException {
            Session session;
            synchronized (sessions) {
                session = getAvailable();
                if (session == null) {
                    logger.log(BasicLevel.ERROR, "no available session, stop executing");
                    return STOP_EXECUTING;
                }
                if (session.closed) {
                    if (!sessions.remove(session)) {
                        Error.bug(logger);
                    }
                    return STOP_EXECUTING;
                }
            }
            session.receiveMessage();
            synchronized (sessions) {
                setAvailable(session);
            }
            return EXECUTE_AGAIN;
        }

        /**
         * @see
         *  org.objectweb.dream.control.activity.task.Task#registered(Object)
         */
        public void registered(Object controlItf) {
            setThreadPoolManager((ThreadPoolManager) controlItf);
        }
    }

    // ------------------------------------------------------------------------
    // ---
    // Utility methods
    // ------------------------------------------------------------------------
    // ---

    protected void setAvailable(Session session) {
        if (firstAvailable == null) {
            if (DEFENSIVE_CHECK && lastAvailable != null) {
                Error.bug(logger);
            }
            firstAvailable = session;
            lastAvailable = session;
        } else {
            if ((DEFENSIVE_CHECK)
                    && (session.nextAvailable != null || lastAvailable.nextAvailable != null)) {
                Error.bug(logger);
            }
            lastAvailable.nextAvailable = session;
            lastAvailable = session;
        }
    }

    protected Session getAvailable() {
        if (firstAvailable == null) {
            if (DEFENSIVE_CHECK && lastAvailable != null) {
                Error.bug(logger);
            }
            return null;
        }
        Session available = firstAvailable;
        firstAvailable = available.nextAvailable;
        if (firstAvailable == null) {
            lastAvailable = null;
        }
        available.nextAvailable = null;
        return available;
    }

    protected void sessionError(Session session, Exception exception) {
        synchronized (sessions) {
            sessions.remove(session);
        }
        silentClose(session);
        if (session.upperIncomingPushItf != null) {
            // session is a server session, inform channel.
            session.upperIncomingPushItf.incomingClosed(session, exception);
        }
    }

    protected void initSession(Session session) {
        synchronized (sessions) {
            sessions.add(session);
            setAvailable(session);
            if (session.upperIncomingPushItf != null && threadPoolManager != null) {
                try {
                    logger.log(BasicLevel.DEBUG, "add a thread in thread pool");
                    threadPoolManager.addThread(readerTask);
                } catch (ThreadPoolOverflowException e) {
                    logger.log(BasicLevel.WARN, "Unable to add reader thread", e);
                } catch (IllegalTaskException e) {
                    Error.bug(logger, e);
                }
            }
        }
    }

    protected void setThreadPoolManager(ThreadPoolManager threadPoolManager) {
        this.threadPoolManager = threadPoolManager;
        synchronized (sessions) {
            Iterator<Session> iterator = sessions.iterator();
            while (iterator.hasNext()) {
                Session session = iterator.next();
                if (session.upperIncomingPushItf == null) {
                    continue;
                }
                try {
                    logger.log(BasicLevel.DEBUG, "add a thread in thread pool");
                    threadPoolManager.addThread(readerTask);
                } catch (ThreadPoolOverflowException e) {
                    logger.log(BasicLevel.WARN, "Unable to add reader thread", e);
                } catch (IllegalTaskException e) {
                    Error.bug(logger, e);
                }
            }
        }
    }

    // -------------------------------------------------------------------------
    // Overridden methods
    // -------------------------------------------------------------------------

    /**
     * @see AbstractComponent#beforeFirstStart(Component)
     */
    @DreamLifeCycle(on = DreamLifeCycleType.FIRST_START)
    protected void beforeFirstStart(Component componentItf) throws IllegalLifeCycleException {
        try {
            Map<String, Object> hints = new HashMap<String, Object>();
            hints.put("thread", "pool");
            Util.addTask(componentItf, readerTask, hints);
        } catch (Exception e) {
            throw new IllegalLifeCycleException("Can't add task");
        }
    }

    /**
     * @see
     *  org.objectweb.dream.control.lifecycle.PrepareStopLifeCycleController#
     *  prepareStopFc()
     */
    @DreamLifeCycle(on = DreamLifeCycleType.PREPARE_TO_STOP)
    public void prepareStopFc() throws IllegalLifeCycleException {
        synchronized (sessions) {
            logger.log(BasicLevel.DEBUG, "Going to stop, close every opened sockets");
            for (Session s : sessions) {
                logger.log(BasicLevel.DEBUG, "Going to close " + s);
                silentClose(s);
            }
        }
    }

    /**
     * Close the session and ignore potential IOException
     * 
     * @param session
     *      the session to close
     */
    protected final void silentClose(Session session) {
        try {
            session.outgoingClose(null);
        } catch (IOException e) {
            // ignore
            logger.log(BasicLevel.DEBUG, "An error occured while closing " + session + " : "
                    + e.getMessage());
        }
    }
}
