/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s):
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing.overChannel;

import java.io.IOException;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.util.Error;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.dream.cache.lib.BasicCacheManagerAttributeController;
import org.objectweb.dream.cache.api.CacheEntry;
import org.objectweb.dream.cache.api.CacheEntryFilter;
import org.objectweb.dream.cache.api.CacheException;
import org.objectweb.dream.cache.api.CacheManager;
import org.objectweb.dream.cache.api.FixableCacheEntry;
import org.objectweb.dream.cache.api.UnFixProtocolException;
import org.objectweb.dream.cache.api.UnbindManager;
import org.objectweb.dream.cache.replacement.api.ReplaceableCacheEntry;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Identified Session (i.e. with a remote export identifier) Manager
 * 
 * @author pgarcia
 */
public class IdentifiedSessionManager {
    private static final boolean DEFENSIVE_CHECKS = false;

    private static final Session NULL_SESSION = new Session();;

    /**
     * Logger of the class
     */
    Logger logger;

    /**
     * Cache fractal interface to bind/fix cache entries
     */
    private CacheManager cacheManagerItf;

    /**
     * Cache fractal interface to unbind cache entries
     */
    private UnbindManager unbindManagerItf;

    /**
     * Cache entry filter for the closeAll() method
     */
    private CacheEntryFilter cacheEntryFilterClosingSocket = new CacheEntryFilterClosingSocket();

    /**
     * Cache entry filter to use only unfixed entries
     */
    private CacheEntryFilter cacheEntryfilter = new CacheUnfixedEntryFilter();

    /**
     * Control interface used to manage parameters of the cache like the
     * maxSize, the currentSize...
     */
    protected BasicCacheManagerAttributeController cacheAttributeControllerItf;

    /**
     * Constructor
     * 
     * @param messagePassingOverChannelImpl
     *      : the related protocol
     */
    public IdentifiedSessionManager(MessagePassingOverChannelImpl messagePassingOverChannelImpl) {
        this.logger = messagePassingOverChannelImpl.logger;
        this.cacheManagerItf = messagePassingOverChannelImpl.cacheManagerItf;
        this.unbindManagerItf = messagePassingOverChannelImpl.unbindManagerItf;
        this.cacheAttributeControllerItf = messagePassingOverChannelImpl.cacheAttributeControllerItf;
    }

    /**
     * @param to
     * @param ce
     * @return
     * @throws IOPushException
     */
    public void reservSession(Session session) throws IOPushException {
        synchronized (session) {
            if (isInUse(session)) {
                getSessionInUse(session);
            } else {
                FixableCacheEntry ce = (FixableCacheEntry) this.cacheManagerItf
                        .lookup(session.remoteExportIdentifier);
                try {
                    this.cacheManagerItf.fix(ce);
                } catch (CacheException e) {
                    Error.bug(this.logger, e);
                }
            }
            if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                this.logger.log(BasicLevel.DEBUG, "Session ok for "
                        + session.remoteExportIdentifier);
            }
        }
    }

    /**
     * @param to
     * @param session
     * @param ce
     * @throws IOPushException
     */
    private void getSessionInUse(Session session) throws IOPushException {
        FixableCacheEntry ce = (FixableCacheEntry) cacheManagerItf
                .lookup(session.remoteExportIdentifier);
        // session found in cache but is in use.
        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "The session found in cache for "
                    + session.remoteExportIdentifier + " is in use : " + ce.getCeFixCount()
                    + " users. Object : " + ce);
        }
        try {
            session.wait();
        } catch (InterruptedException e) {
            throw new IOPushException("Interrupted while waiting for the session toward "
                    + session.remoteExportIdentifier + " to be available");
        }

        try {
            this.cacheManagerItf.fix(ce);
        } catch (CacheException e) {
            Error.bug(this.logger, e);
        }
    }

    /**
     * Close all sessions in cache
     */
    public synchronized void closeAll() {
        LoggerUtil.start(logger);
        try {
            unbindManagerItf.unbind(cacheEntryFilterClosingSocket, true);
        } catch (CacheException e) {
            Error.bug(this.logger, e);
        }
        LoggerUtil.end(logger);
    }

    /**
     * Bind the export identifier to the related session
     * 
     * @param to
     * @param newSession
     * @throws IOPushException
     */
    public void bind(ExportIdentifier to, Session newSession) {
        // open new session.
        try {
            waitForCacheSpace();
        } catch (CacheException e) {
            Error.bug(logger, e);
        }
        try {
            if (DEFENSIVE_CHECKS) {
                if (newSession == null)
                    Error.bug(logger);
                if (to == null)
                    Error.bug(logger);
                CacheEntry ce = this.cacheManagerItf.lookup(to);
                if (ce != null) {
                    if (ce.getCeObject() != null) {
                        Session sessionInCache = (Session) ce.getCeObject();
                        if (!sessionInCache.equals(NULL_SESSION)) {
                            if (sessionInCache.equals(newSession)) {
                                logger.log(BasicLevel.ERROR, sessionInCache
                                        + " is already in cache");
                            } else {
                                logger.log(BasicLevel.ERROR, newSession + " cannot replace "
                                        + sessionInCache);
                            }
                            Error.bug(logger);
                        }
                    }
                }
            }
            this.cacheManagerItf.bind(to, newSession);
        } catch (CacheException e) {
            Error.bug(this.logger, e);
        }
    }

    /**
     * Unbind the entry.
     * 
     * @param id
     *      the related export identifier
     * @return false if the entry was really unbound
     */
    public boolean unbind(ExportIdentifier id) {
        if (DEFENSIVE_CHECKS && (id == null)) {
            Error.bug(this.logger);
        }
        try {
            if (isInUse(id))
                unfix(getSession(id));
            boolean res = unbindManagerItf.unbind(id, false);
            cacheManagerItf.bind(id, NULL_SESSION);
            unbindManagerItf.unbind(id, false);
            synchronized (this.cacheAttributeControllerItf) {
                cacheAttributeControllerItf.notify();
            }
            return res;
        } catch (CacheException e) {
            Error.bug(logger, e);
        }
        return false;
    }

    /**
     * Mark the session as used
     * 
     * @param cacheEntry
     */
    public void fix(Session session) {
        FixableCacheEntry cacheEntry = (FixableCacheEntry) cacheManagerItf
                .lookup(session.remoteExportIdentifier);
        if (DEFENSIVE_CHECKS && cacheEntry == null)
            Error.bug(logger);
        cacheEntry.fixCe();
    }

    /**
     * Mark the session as unused
     * 
     * @param cacheEntry
     */
    public void unfix(Session session) {
        try {
            FixableCacheEntry cacheEntry = (FixableCacheEntry) cacheManagerItf
                    .lookup(session.remoteExportIdentifier);
            if (DEFENSIVE_CHECKS && cacheEntry == null)
                Error.bug(logger);
            if (!isInUse(cacheEntry)) {
                return;
            }
            cacheEntry.unfixCe();
            if (this.logger.isLoggable(BasicLevel.DEBUG) && isInUse(cacheEntry)) {
                Error.bug(logger);
            }
            synchronized (this.cacheAttributeControllerItf) {
                cacheAttributeControllerItf.notify();
            }
        } catch (UnFixProtocolException e) {
            Error.bug(logger, e);
        }
    }

    /**
     * Lookup for a session identifier by the given export identifier
     * 
     * @param to
     * @return
     */
    public Session lookup(ExportIdentifier to) {
        ReplaceableCacheEntry ce = (ReplaceableCacheEntry) cacheManagerItf.lookup(to);
        if (ce != null) {
            Session s = (Session) ce.getCeObject();
            if (!s.equals(NULL_SESSION)) {
                return s;
            }
        }
        return null;
    }

    /**
     * Wait for cache space
     * 
     * @throws BindException
     */
    public void waitForCacheSpace() throws CacheException {
        synchronized (this.cacheAttributeControllerItf) {
            while (isCacheFull()) {
                this.logger.log(BasicLevel.INFO, "Cache of 'messagePassing' sessions is full");

                // try to remove unused sessions
                try {
                    this.unbindManagerItf.unbind(this.cacheEntryfilter, false);
                } catch (CacheException e1) {
                    logger.log(BasicLevel.WARN,
                            "an interruption occured while trying to remove unused session");
                }

                try {
                    // wait for 1 second if a session is released
                    logger.log(BasicLevel.DEBUG, "Waiting for cache space");
                    this.cacheAttributeControllerItf.wait(1000);
                } catch (InterruptedException e) {
                    logger.log(BasicLevel.WARN, "Interrupted while waiting for cache space");
                }
            }
            if (isCacheFull()) {
                // no session released
                throw new CacheException("Cache of messagePassing sessions (size="
                        + cacheAttributeControllerItf.getCacheMaxSize() + ") is full");
            }
        }
    }

    /**
     * Filter used to close every session in cache
     */
    private class CacheEntryFilterClosingSocket implements CacheEntryFilter {
        /**
         * Close session contained in every cache entry
         * 
         * @return <code>true</code>
         */
        public boolean accept(CacheEntry ce) {
            logger.log(BasicLevel.DEBUG, "Closing session : " + ce.getCeIdentifier());
            Session s = (Session) ce.getCeObject();
            if (!s.equals(NULL_SESSION)) {
                try {
                    s.close();
                } catch (IOException e) {
                    logger.log(BasicLevel.WARN, "An error occurs while closing session", e);
                }
            } else {
                logger.log(BasicLevel.DEBUG, "the session for " + ce.getCeIdentifier()
                        + "has been unbound");
            }
            return true;
        }
    }

    /**
     * Filter used to treat only unused sessions
     */
    private class CacheUnfixedEntryFilter implements CacheEntryFilter {

        public boolean accept(CacheEntry arg0) {
            return ((FixableCacheEntry) arg0).getCeFixCount() == 0;
        }
    }

    /**
     * Get the session contained in a cache entry
     * 
     * @param ce
     * @return
     */
    Session getSession(ExportIdentifier id) {
        return (Session) cacheManagerItf.lookup(id).getCeObject();
    }

    boolean isInUse(FixableCacheEntry ce) {
        return ce.getCeFixCount() > 0;
    }

    /**
     * @return true if the cache is full
     */
    private boolean isCacheFull() {
        return this.cacheAttributeControllerItf.getCurrentSize() >= this.cacheAttributeControllerItf
                .getCacheMaxSize();
    }

    /**
     * @param session
     * @return
     */
    boolean isInUse(Session session) {
        return (((FixableCacheEntry) cacheManagerItf.lookup(session.remoteExportIdentifier))
                .getCeFixCount()) > 0;
    }

    /**
     * @param id
     * @return
     */
    boolean isInUse(ExportIdentifier id) {
        return (((FixableCacheEntry) cacheManagerItf.lookup(id)).getCeFixCount()) > 0;
    }
}
