/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Pierre Garcia
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.messagePassing.overChannel;

import java.io.IOException;

import org.objectweb.dream.IOPushException;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.BindException;
import org.objectweb.dream.protocol.ExportException;
import org.objectweb.dream.protocol.ExportIdentifier;
import org.objectweb.dream.protocol.ExportIdentifierChunk;
import org.objectweb.dream.protocol.IncomingPush;
import org.objectweb.dream.protocol.InvalidExportIdentifierException;
import org.objectweb.dream.protocol.OutgoingPush;
import org.objectweb.dream.protocol.channel.ChannelFactory;
import org.objectweb.dream.protocol.channel.ChannelProtocol;
import org.objectweb.dream.protocol.channel.CloseChunk;
import org.objectweb.dream.protocol.messagePassing.MessagePassingOutgoingPush;
import org.objectweb.dream.util.Error;
import org.objectweb.dream.util.LoggerUtil;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

class SessionManager implements MessagePassingOutgoingPush, ChannelFactory {
    /**
     * for debug information. The code should optimize the code when false
     */
    private static final boolean DEFENSIVE_CHECKS = false;

    /**
     * Inherited Logger of the protocol component
     */
    protected Logger logger;

    /**
     * Interface for the management of dream messages.
     */
    protected MessageManagerType messageManagerItf;

    /**
     * Interface related to the export/bind pattern.
     */
    protected ChannelProtocol lowerProtocolItf;

    /**
     * Chunk factory to create a chunk containing an {@link ExportIdentifier}
     */
    protected ChunkFactoryReference<ExportIdentifierChunk> exportIdChunkFactory;

    /**
     * Chunk factory to send close request to a session
     */
    protected ChunkFactoryReference<CloseChunk> closeChunkFactory;

    /**
     * Manager for anonymous session
     */
    protected AnonymousSessionManager anonymousSessionManager;

    /**
     * Manager for identified session, i.e. linked to an export identifier
     */
    protected IdentifiedSessionManager identifiedSessionManager;

    /**
     * Boolean variable used to know the state of the access point
     * <ul>
     * <li><code>true</code> if the communication channel (interface) has been
     * exported</li>
     * <li><code>false</code> otherwise</li>
     * </ul>
     */
    protected Boolean closed = false;

    protected String fromChunkName;

    protected ExportIdentifier localExportIdentifier;

    /**
     * Upper incoming push to transmit message upper in the protocol stack
     */
    protected IncomingPush upperIncomingPushItf;

    /**
     * Default constructor
     * 
     * @param messagePassingOverChannelImpl
     * @param upperIncomingPushItf
     * @param fromChunkName
     */
    SessionManager(MessagePassingOverChannelImpl messagePassingOverChannelImpl,
            IncomingPush upperIncomingPushItf, String fromChunkName) {
        // inherited references
        this.logger = messagePassingOverChannelImpl.logger;
        this.messageManagerItf = messagePassingOverChannelImpl.messageManagerItf;
        this.lowerProtocolItf = messagePassingOverChannelImpl.lowerProtocolItf;

        this.anonymousSessionManager = new AnonymousSessionManager(
                messagePassingOverChannelImpl.logger);
        this.identifiedSessionManager = new IdentifiedSessionManager(messagePassingOverChannelImpl);

        this.upperIncomingPushItf = upperIncomingPushItf;
        this.fromChunkName = fromChunkName;
    }

    // -------------------------------------------------------------------------
    // Implementation of the ChannelFactory interface
    // -------------------------------------------------------------------------

    /**
     * @see ChannelFactory#instantiate(OutgoingPush)
     */
    public IncomingPush instantiate(OutgoingPush toClientPush) throws BindException {
        LoggerUtil.start(this.logger);
        this.logger.log(BasicLevel.DEBUG, "New incoming connection to instantiate");
        if (DEFENSIVE_CHECKS && toClientPush == null) {
            Error.bug(this.logger);
        }

        if (closed) {
            throw new BindException("Access point has been closed");
        }
        this.logger.log(BasicLevel.DEBUG, "New anonymous session ");

        Session session = new Session(toClientPush, this);
        anonymousSessionManager.add(session);
        LoggerUtil.returnAfter(this.logger);
        return session;
    }

    // -------------------------------------------------------------------------
    // Implementation of the MessagePassingOutgoingPush interface
    // -------------------------------------------------------------------------

    /**
     * @see MessagePassingOutgoingPush#getLocalExportIdentifier()
     */
    public ExportIdentifier getLocalExportIdentifier() {
        return localExportIdentifier;
    }

    /**
     * @see MessagePassingOutgoingPush#outgoingPush(Message, ExportIdentifier)
     */
    public void outgoingPush(Message message, ExportIdentifier to)
            throws InvalidExportIdentifierException, IOPushException {
        LoggerUtil.start(this.logger);
        if (localExportIdentifier != null && localExportIdentifier.equals(to)) {
            // local sending : the message is directly pushed up in the stack
            if (fromChunkName != null) {
                ExportIdentifierChunk chunk = newExportIdentifierChunk();
                chunk.setExportIdentifier(localExportIdentifier);
                messageManagerItf.addChunk(message, fromChunkName, chunk);
            }
            try {
                upperIncomingPushItf.incomingPush(message);
            } catch (PushException e) {
                throw new IOPushException(e.getMessage());
            }
            return;
        }
        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "Remote send : export identifier in the message ("
                    + to + "), local is (" + localExportIdentifier + ")");
        }

        synchronized (identifiedSessionManager) {
            // look for an identified session in the related manager
            if (closed) {
                throw new IOPushException("Access point has been closed.");
            }
            if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                this.logger.log(BasicLevel.DEBUG, "Try to get Session in cache for " + to);
            }

            Session session = identifiedSessionManager.lookup(to);
            if (session == null) {
                // session not found in cache, bind and create new session
                session = createAndBindNewSession(to);
            } else {
                // get the session
                if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                    this.logger.log(BasicLevel.DEBUG, "Session found in cache for " + to);
                }
                identifiedSessionManager.reservSession(session);
            }

            // push outgoing message
            IOPushException exception = null;
            try {
                session.outgoingPush.outgoingPush(message);
            } catch (IOPushException e) {
                exception = e;
            }
            // synchronized (session)
            // {
            if (this.logger.isLoggable(BasicLevel.DEBUG)) {
                this.logger.log(BasicLevel.DEBUG, "Release socket (error=" + exception + ") for "
                        + to);
            }
            Session replacingSession = releaseSession(session, exception);
            if (exception != null) {
                try {
                    if (replacingSession == null) {
                        // the session must be recreated
                        tryToRePush(to, message);
                    } else {
                        // try to send the session via the replacing session
                        if (DEFENSIVE_CHECKS
                                && (session.remoteExportIdentifier != replacingSession.remoteExportIdentifier)) {
                            Error.bug(this.logger);
                        }
                        identifiedSessionManager.fix(replacingSession);
                        tryToRePush(replacingSession, message);
                    }
                } catch (IOPushException e) {
                    logger.log(BasicLevel.INFO, "An error occured while resending the message : "
                            + e.getMessage());
                    throw exception;
                }
            }
        }

        LoggerUtil.end(this.logger);
    }

    /**
     * Try to push the message again via the replacing session
     * 
     * @param replacingSession
     *      the replacing session
     * @param message
     *      the message to resent
     * @throws IOPushException
     */
    private void tryToRePush(Session replacingSession, Message message) throws IOPushException {
        // resend message
        try {
            replacingSession.outgoingPush.outgoingPush(message);
        } catch (IOPushException e) {
            identifiedSessionManager.unfix(replacingSession);
            throw new IOPushException("the message cannot be sent towards "
                    + replacingSession.remoteExportIdentifier);
        }
        identifiedSessionManager.unfix(replacingSession);
    }

    /**
     * Try to send again the message toward a destination :
     * <ol>
     * <li>Recreate and bind a new session</li>
     * <li>Try to send the message via this session</li>
     * </ol>
     * 
     * @param to
     *      the export identifier of the remote destination
     * @param message
     *      the message to resent
     * @throws IOPushException
     */
    private void tryToRePush(ExportIdentifier to, Message message) throws IOPushException {
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "Try to push again the message towards " + to);
        }
        // recreate session
        Session newSession = null;
        try {
            newSession = createAndBindNewSession(to);
        } catch (IOPushException e) {
            throw new IOPushException("the session cannot be recreated towards " + to);
        } catch (InvalidExportIdentifierException e) {
            Error.bug(logger);
        }
        tryToRePush(newSession, message);
    }

    /**
     * Create/Use an entry in the cache and proceed the bind step of the
     * export/bind pattern
     * 
     * @param to
     * @return
     * @throws InvalidExportIdentifierException
     * @throws IOPushException
     */
    private Session createAndBindNewSession(ExportIdentifier to)
            throws InvalidExportIdentifierException, IOPushException {
        Session newSession = null;

        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "Create new session to " + to);
        }

        newSession = new Session(to, this);
        OutgoingPush lowerOutgoingPush = null;
        try {
            lowerOutgoingPush = bindSession(to, newSession);
        } catch (BindException e) {
            throw new IOPushException(e);
        }

        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger.log(BasicLevel.DEBUG, "Bind step of export/bind pattern to " + to
                    + " done.");
        }

        if (DEFENSIVE_CHECKS && lowerOutgoingPush == null) {
            Error.bug(this.logger);
        }
        // synchronized (identifiedSessionManager)
        // {
        identifiedSessionManager.bind(to, newSession);
        identifiedSessionManager.fix(newSession);

        newSession.init(lowerOutgoingPush);
        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.logger
                    .log(BasicLevel.DEBUG, "New session toward (" + to + ") completely created.");
            // }
        }
        return newSession;
    }

    /**
     * Release the session and try to replace it.
     * 
     * @param sessionToRelease
     *      the session to release
     * @param exception
     *      the potential exception which occurred during the sent
     * @return the replacing session if it exist, <code>null</code> otherwise
     * @throws IOPushException
     *      if an error occurs while releasing the session
     */
    private Session releaseSession(Session sessionToRelease, IOPushException exception)
            throws IOPushException {
        LoggerUtil.start(logger);
        // re-put session in cache.
        if (DEFENSIVE_CHECKS && (sessionToRelease == null)) {
            Error.bug(this.logger);
        }
        // if (DEFENSIVE_CHECKS
        // && (!identifiedSessionManager.isInUse(sessionToRelease)))
        // {
        // logger.log(BasicLevel.ERROR, sessionToRelease +" that will be release
        // is
        // not in use " + (exception != null ? exception.getMessage() : ""));
        // Error.bug(this.logger);
        // }
        if (exception != null) {
            this.logger.log(BasicLevel.DEBUG, "An exception occurs while sending the message.",
                    exception);
            LoggerUtil.returnAfter(logger);
            return tryReplaceSessionAfterException(sessionToRelease);
        } else {
            LoggerUtil.returnAfter(logger);
            return replaceSessionNotInUse(sessionToRelease);
        }
    }

    /**
     * Replace the session :
     * <ol>
     * <li>Send a close request</li>
     * <li>Unbind in cache the old session</li>
     * <li>Bind in cache the new session</li>
     * </ol>
     * 
     * @param ce
     * @throws IOPushException
     */
    private Session replaceSessionNotInUse(Session sessionToReplace) throws IOPushException {
        if (sessionToReplace.replacingSession != null) {
            this.logger.log(BasicLevel.DEBUG, "Session has to be replaced by "
                    + "another one. Send close request and replace it.");

            // send the close request
            sessionToReplace.closeRequestSent = true;
            Message closeMessage = this.messageManagerItf.createMessage();
            this.messageManagerItf.addChunk(closeMessage, CloseChunk.DEFAULT_NAME, newCloseChunk());
            try {
                sessionToReplace.outgoingPush.outgoingPush(closeMessage);
            } catch (IOPushException e) {
                this.logger
                        .log(BasicLevel.WARN, "exception caught while sending close message.", e);
            }

            // unbind the entry in the cache
            identifiedSessionManager.unbind(sessionToReplace.remoteExportIdentifier);

            // bind the new session in cache
            identifiedSessionManager.bind(sessionToReplace.remoteExportIdentifier,
                    sessionToReplace.replacingSession);
        } else {
            identifiedSessionManager.unfix(sessionToReplace);
        }
        synchronized (sessionToReplace) {
            if (!identifiedSessionManager.isInUse(sessionToReplace)) {
                sessionToReplace.notify();
            }
        }
        return sessionToReplace.replacingSession;
    }

    /**
     * try to replace the given session
     * 
     * @param sessionToReplace
     *      the session that will be replaced
     * @return <ul>
     *  <li>The replacing {@link Session} if the session was really replaced
     *  </li>
     *  <li>null if the {@link Session} was simply removed</li>
     *  </ul>
     */
    private Session tryReplaceSessionAfterException(Session sessionToReplace) {
        LoggerUtil.start(logger);
        try {
            sessionToReplace.outgoingPush.outgoingClose(sessionToReplace);
        } catch (IOException e) {
            // ignore
        }
        if (DEFENSIVE_CHECKS && (sessionToReplace.remoteExportIdentifier == null)) {
            Error.bug(this.logger);
        }
        identifiedSessionManager.unbind(sessionToReplace.remoteExportIdentifier);
        if (sessionToReplace.replacingSession != null) {
            this.logger.log(BasicLevel.DEBUG, "Session has to be replaced by "
                    + "another one. Close and replace it.");

            identifiedSessionManager.bind(sessionToReplace.remoteExportIdentifier,
                    sessionToReplace.replacingSession);
            synchronized (sessionToReplace) {
                if (!identifiedSessionManager.isInUse(sessionToReplace)) {
                    sessionToReplace.notify();
                }
            }
            if (DEFENSIVE_CHECKS
                    && (identifiedSessionManager.lookup(sessionToReplace.remoteExportIdentifier) == null)) {
                Error.bug(this.logger);
            }
            LoggerUtil.returnAfter(logger);
            return sessionToReplace.replacingSession;
        } else {
            if (DEFENSIVE_CHECKS
                    && (identifiedSessionManager.lookup(sessionToReplace.remoteExportIdentifier) != null)) {
                Error.bug(this.logger);
            }
            LoggerUtil.returnAfter(logger);
            return null;
        }

    }

    /**
     * Bind the session. bind represents here the bind step of the export/bind
     * pattern
     * 
     * @param to
     * @param ce
     * @return
     * @throws InvalidExportIdentifierException
     * @throws BindException
     */
    private OutgoingPush bindSession(ExportIdentifier to, Session session)
            throws InvalidExportIdentifierException, BindException {
        LoggerUtil.start(this.logger);
        OutgoingPush lowerOutgoingPush;
        if (upperIncomingPushItf == null) {
            this.logger
                    .log(BasicLevel.DEBUG, "Bind call on lower protocol : no upperIncomingPush ");
            lowerOutgoingPush = this.lowerProtocolItf.bind(to, null, null);
        } else {
            lowerOutgoingPush = this.lowerProtocolItf.bind(to, session, null);
        }
        LoggerUtil.returnAfter(this.logger);
        return lowerOutgoingPush;
    }

    /**
     * @see MessagePassingOutgoingPush#outgoingClose(IncomingPush)
     */
    public void outgoingClose(IncomingPush incomingPush) throws IOException {
        LoggerUtil.start(this.logger);
        synchronized (identifiedSessionManager) {
            closed = true;
            closeLocalAccessPoint();
            identifiedSessionManager.closeAll();
            anonymousSessionManager.closeAll();
        }
        LoggerUtil.end(this.logger);
    }

    private void closeLocalAccessPoint() throws ExportException {
        this.logger.log(BasicLevel.DEBUG, "Close local access point " + localExportIdentifier);
        if (localExportIdentifier != null) {
            try {
                this.logger.log(BasicLevel.DEBUG, "Unexport lower channel factory.");
                lowerProtocolItf.unexport(localExportIdentifier);
            } catch (InvalidExportIdentifierException e) {
                throw new ExportException("Unable to unexport previously exported channel factory",
                        e);
            }
        }
    }

    /**
     * Create an {@link ExportIdentifierChunk}
     * 
     * @return an empty export identifier chunk
     */
    private ExportIdentifierChunk newExportIdentifierChunk() {
        if (exportIdChunkFactory == null) {
            exportIdChunkFactory = messageManagerItf.getChunkFactory(ExportIdentifierChunk.class);
        }
        return messageManagerItf.createChunk(exportIdChunkFactory);
    }

    /**
     * Create a {@link CloseChunk}
     * 
     * @return
     */
    private CloseChunk newCloseChunk() {
        if (closeChunkFactory == null) {
            closeChunkFactory = messageManagerItf.getChunkFactory(CloseChunk.class);
        }
        return messageManagerItf.createChunk(closeChunkFactory);
    }

    /**
     * Remove anonymous session delegate the call to the {@link
     * AnonymousSessionManager}.
     * 
     * @param session
     *      the session to remove
     */
    public final void removeAnonymousSession(Session session) {
        synchronized (identifiedSessionManager) {
            Session sessionInCache = identifiedSessionManager
                    .lookup(session.remoteExportIdentifier);
            // close session request
            if (DEFENSIVE_CHECKS && sessionInCache != null) {
                Error.bug(this.logger);
            }
            anonymousSessionManager.remove(session);
        }
    }

    /**
     * Remove a {@link Session} identified by a {@link ExportIdentifier}.
     * 
     * @param remoteExportIdentifier
     *      the identifier of the session. if <code>null</code>, the session is
     *      considered as anonymous.
     * @param sessionToRemove
     */
    public final void removeSession(ExportIdentifier remoteExportIdentifier, Session sessionToRemove) {
        synchronized (identifiedSessionManager) {
            LoggerUtil.start(logger);
            Session sessionInCache = identifiedSessionManager.lookup(remoteExportIdentifier);
            if (sessionInCache != null) {
                identifiedSessionManager.unbind(remoteExportIdentifier);
                if (this.logger.isLoggable(BasicLevel.INFO)) {
                    this.logger.log(BasicLevel.INFO, sessionToRemove
                            + " is removed from the cache ");
                }
            } else {
                synchronized (anonymousSessionManager) {
                    if (anonymousSessionManager.contains(sessionToRemove))
                        removeAnonymousSession(sessionToRemove);
                }
            }
            LoggerUtil.end(logger);
        }
    }

    /**
     * @param session
     * @throws IOPushException
     */
    public void sessionIdentified(ExportIdentifier exportIdentifier, Session session)
            throws IOPushException {
        synchronized (identifiedSessionManager) {
            Session sessionInCache = identifiedSessionManager
                    .lookup(session.remoteExportIdentifier);
            if (sessionInCache == null) {
                newRemoteDestination(session.remoteExportIdentifier, session);
            } else {
                this.logger.log(BasicLevel.DEBUG, "Remote destination ("
                        + session.remoteExportIdentifier + ") already exist");

                // remote destination already exist
                if ((localExportIdentifier.hashCode() != session.remoteExportIdentifier.hashCode())
                        && sessionInCache.remoteExportIdentifier == null) {
                    replaceSession(sessionInCache, session);
                }
            }
        }
    }

    /**
     * Replace session
     * 
     * @param oldSession
     *      the session to replace
     * @param newSession
     *      the new session
     * @throws IOPushException
     */
    private void replaceSession(Session oldSession, Session newSession) throws IOPushException {
        LoggerUtil.start(this.logger);
        this.logger.log(BasicLevel.DEBUG, "Close session in cache and replace it by this one");
        anonymousSessionManager.remove(newSession);

        // close the session already in cache :
        // first wait until it is free
        // then if the session is still opened, send a close
        // message
        // finally replace by the new one in the cache entry
        synchronized (oldSession) {
            if (identifiedSessionManager.isInUse(oldSession)) {
                this.logger.log(BasicLevel.DEBUG, "Session in cache is in use,"
                        + " will be closed and replaced later");
                oldSession.replacingSession = newSession;
            } else {
                oldSession.replacingSession = newSession;
                replaceSessionNotInUse(oldSession);
            }
        }
        LoggerUtil.end(this.logger);
    }

    /**
     * Mark a session as identified
     * 
     * @param exportIdentifier
     *      the identifier of the session
     * @throws IOPushException
     *      if an error occurred while identifying the session
     */
    private void newRemoteDestination(ExportIdentifier exportIdentifier, Session session)
            throws IOPushException {
        LoggerUtil.start(this.logger);
        this.logger.log(BasicLevel.DEBUG, "New remote destination");
        removeAnonymousSession(session);
        session.remoteExportIdentifier = exportIdentifier;

        // remote destination was unknown
        identifiedSessionManager.bind(session.remoteExportIdentifier, session);
        LoggerUtil.end(this.logger);
    }
}
