/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.rpc;

import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import org.objectweb.dream.message.AbstractChunk;

public class ReplyChunk extends AbstractChunk<ReplyChunk>
{

  /**
   * Automatically generated
   */
  private static final long  serialVersionUID = 3453406932310868482L;

  /** The default name of chunks of this type. */
  public static final String DEFAULT_NAME     = "reply_chunk";

  long                       invocationId;
  Object                     reply;
  Throwable                  throwable;

  // ---------------------------------------------------------------------------
  // Chunk methods
  // ---------------------------------------------------------------------------

  /** Default constructor. */
  public ReplyChunk()
  {
  }

  /**
   * @return Returns the invocationId.
   */
  public long getInvocationId()
  {
    return invocationId;
  }

  /**
   * @param invocationId The invocationId to set.
   */
  public void setInvocationId(long invocationId)
  {
    this.invocationId = invocationId;
  }

  /**
   * @return Returns the reply.
   */
  public Object getReply()
  {
    return reply;
  }

  /**
   * @param reply The reply to set.
   */
  public void setReply(Object reply)
  {
    if (reply != null && throwable != null)
    {
      throw new IllegalArgumentException(
          "throwable and reply cannot be non null at the same time.");
    }
    this.reply = reply;
  }

  /**
   * @return Returns the throwable.
   */
  public Throwable getThrowable()
  {
    return throwable;
  }

  /**
   * @param throwable The throwable to set.
   */
  public void setThrowable(Throwable throwable)
  {
    if (reply != null && throwable != null)
    {
      throw new IllegalArgumentException(
          "throwable and reply cannot be non null at the same time.");
    }
    this.throwable = throwable;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Chunk interface
  // ---------------------------------------------------------------------------

  /**
   * @see AbstractChunk#newChunk()
   */
  public ReplyChunk newChunk()
  {
    return new ReplyChunk();
  }

  /**
   * @see AbstractChunk#transfertStateTo(AbstractChunk)
   */
  public void transfertStateTo(ReplyChunk chunk)
  {
    chunk.invocationId = invocationId;
    chunk.reply = reply;
    chunk.throwable = throwable;
  }

  /**
   * @see org.objectweb.dream.pool.Recyclable#recycle()
   */
  public void recycle()
  {
    invocationId = 0;
    reply = null;
    throwable = null;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Externalizable interface
  // ---------------------------------------------------------------------------

  /**
   * @see java.io.Externalizable#readExternal(ObjectInput)
   */
  public void readExternal(ObjectInput in) throws IOException,
      ClassNotFoundException
  {
    invocationId = in.readLong();
    boolean b = in.readBoolean();
    if (b)
    {
      // reply object
      reply = in.readObject();
      throwable = null;
    }
    else
    {
      if (in.readBoolean())
      {
        // exception thrown
        reply = null;
        throwable = (Throwable) in.readObject();
      }
      else
      {
        // void reply
        reply = null;
        throwable = null;
      }
    }
  }

  /**
   * @see java.io.Externalizable#writeExternal(java.io.ObjectOutput)
   */
  public void writeExternal(ObjectOutput out) throws IOException
  {
    out.writeLong(invocationId);
    if (reply != null)
    {
      // reply object
      out.writeBoolean(true);
      out.writeObject(reply);
    }
    else
    {
      out.writeBoolean(false);
      if (throwable == null)
      {
        // void reply
        out.writeBoolean(false);
      }
      else
      {
        // exception thrown
        out.writeBoolean(true);
        out.writeObject(throwable);
      }
    }
  }

}
