/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): Romain Lenglet
 */

package org.objectweb.dream.serializator;

import java.io.IOException;


import org.objectweb.dream.IOPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.AbstractChunk;
import org.objectweb.dream.message.ChunkFactoryReference;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.fractal.fraclet.annotation.annotations.Attribute;
import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;

/**
 * This component serializes messages it receives on its input and send them on
 * its output. Serialized messages are messages that contain a
 * {@link ByteArrayChunk} chunk. This chunk is registered under its default
 * name.
 * <p>
 * It is possible to specify the name of a chunk that must not been serialized
 * (it is added into the outgoing message).
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
@Provides(interfaces = @Interface(name = "in-push", signature = Push.class))
public class PushPushSerializatorImpl  implements Push
{

  ChunkFactoryReference<ByteArrayChunk> byteArrayChunkFactory;

  // ---------------------------------------------------------------------------
  // Client interfaces
  // ---------------------------------------------------------------------------

  @Requires(name = SerializatorItf.ITF_NAME)
  protected SerializatorItf<byte []>             serializatorItf;
  @Requires(name = "message-manager")
  protected MessageManagerType          messageManagerItf;
  @Requires(name = "out-push")
  protected Push                        outPushItf;

  // ---------------------------------------------------------------------------
  // Attributes
  // ---------------------------------------------------------------------------
  /**
   * the name of the chunk that must not been serialized.
   */
  @Attribute(argument = "toBeRemovedChunkName")
  protected String                      toBeRemovedChunkName;

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see Push#push(Message)
   */
  public void push(Message message) throws PushException
  {

    AbstractChunk<?> removedChunk = null;
    if (toBeRemovedChunkName != null)
    {
      removedChunk = messageManagerItf.removeChunk(message,
          toBeRemovedChunkName);
    }

    byte[] serializedMessage;
    try
    {
      serializedMessage = serializatorItf.serialize(message);
      messageManagerItf.deleteMessage(message);
    }
    catch (IOException e)
    {
      throw new IOPushException(
          "Unable to serialize the message to be broadcast", e);
    }

    Message toReturn = messageManagerItf.createMessage();
    ByteArrayChunk byteArrayChunk = newByteArrayChunk();
    messageManagerItf.addChunk(toReturn, ByteArrayChunk.DEFAULT_NAME,
        byteArrayChunk);
    byteArrayChunk.setByteArray(serializedMessage);

    if (removedChunk != null)
    {
      messageManagerItf.addChunk(toReturn, toBeRemovedChunkName, removedChunk);
    }
    outPushItf.push(toReturn);

  }

  // ---------------------------------------------------------------------------
  // Utility method
  // ---------------------------------------------------------------------------

  private ByteArrayChunk newByteArrayChunk()
  {
    if (byteArrayChunkFactory == null)
    {
      byteArrayChunkFactory = messageManagerItf
          .getChunkFactory(ByteArrayChunk.class);
    }
    return messageManagerItf.createChunk(byteArrayChunkFactory);
  }
}
