/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.synchro;


import org.objectweb.dream.dreamannotation.DreamComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;

/**
 * ReEntrant implementation of the {@link ConditionItf }interface. This
 * implementation uses a re-entrant mutex.
 * <p>
 * <b>Note: Inspired by Doug Lea's implementation. </b>
 */
@DreamComponent(controllerDesc = "dreamUnstoppablePrimitive")
public class ConditionReEntrantImpl 
    implements
      ConditionItf

{

  // Client interfaces
  @Requires(name=MutexReEntrantItf.ITF_NAME)
  protected MutexReEntrantItf mutexReEntrantItf;

  // ---------------------------------------------------------------------------
  // Implementation of the ConditionItf interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.synchro.ConditionItf#await()
   */
  public void await() throws InterruptedException
  {
    if (Thread.interrupted())
    {
      throw new InterruptedException();
    }
    Object o = null;
    try
    {
      synchronized (this)
      {
        o = mutexReEntrantItf.forceUnlock();
        try
        {
          wait();
        }
        catch (InterruptedException ex)
        {
          notify();
          throw ex;
        }
      }
    }
    finally
    {
      // Must ignore interrupt on re-acquire
      boolean interrupted = false;
      for (;;)
      {
        try
        {
          mutexReEntrantItf.forceLock(o);
          break;
        }
        catch (InterruptedException ex)
        {
          interrupted = true;
        }
      }
      if (interrupted)
      {
        Thread.currentThread().interrupt();
      }
    }
  }

  /**
   * @see org.objectweb.dream.synchro.ConditionItf#timedwait(long)
   */
  public boolean timedwait(long msecs) throws InterruptedException
  {
    if (Thread.interrupted())
      throw new InterruptedException();
    boolean success = false;
    Object o = null;
    try
    {
      synchronized (this)
      {
        o = mutexReEntrantItf.forceUnlock();
        try
        {
          if (msecs > 0)
          {
            long start = System.currentTimeMillis();
            wait(msecs);
            success = System.currentTimeMillis() - start <= msecs;
          }
        }
        catch (InterruptedException ex)
        {
          notify();
          throw ex;
        }
      }
    }
    finally
    {
      // Must ignore interrupt on re-acquire
      boolean interrupted = false;
      for (;;)
      {
        try
        {
          mutexReEntrantItf.forceLock(o);
          break;
        }
        catch (InterruptedException ex)
        {
          interrupted = true;
        }
      }
      if (interrupted)
      {
        Thread.currentThread().interrupt();
      }
    }
    return success;
  }

  /**
   * @see org.objectweb.dream.synchro.ConditionItf#signal()
   */
  public synchronized void signal()
  {
    notify();
  }

  /**
   * @see org.objectweb.dream.synchro.ConditionItf#signalAll()
   */
  public synchronized void signalAll()
  {
    notifyAll();
  }
}